# coding: utf-8
"""
    LinTranAnim.py

    Copyright (c) 2019,2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.patches import Polygon, Circle
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
from molass_legacy.KekLib.OurMatplotlib import get_color
from OurManim import manim_init, use_default_style, Animation, Collection, TextGroup, Parallelogram, Arrow
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar

class LinTranAnim(Dialog):
    def __init__(self, parent=None):
        if parent is None:
            from SerialTestUtils import get_tk_root
            parent = get_tk_root()
        self.parent = parent
        self.num_frames = 70
        manim_init()
        Dialog.__init__(self, parent, "Linear Transformation Tutorial", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        self.fig, self.axes = plt.subplots(nrows=1, ncols=2, figsize=(12,6))
        self.axes[1].set_axis_off()
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, cframe)
        self.toolbar.update()
        self.draw()
        self.fig.tight_layout()
        self.mpl_canvas.draw()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Save", width=10, command=self.save)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

    def draw(self):
        ax1, ax2 = self.axes

        cube_labels = ['(0,0)', '(1,0)', '(1,1)', '(0,1)']
        cube_points = [(0,0), (1,0), (1,1), (0,1)]
        data_labels = ['(a,b)', '(a+c, b+d)', '(c,d)']
        data_points = [(2,0.5), (2.5, 2.0), (0.5, 1.5)]

        labels = cube_labels + data_labels
        points = cube_points + data_points
        d = 0.05

        for t, p in zip(labels, points):
            ax1.plot(*p,'o')
            ax1.text(p[0]+d, p[1]+d, t, fontsize=16)

        ax1.set_xlim(-0.5, 3.0)
        ax1.set_ylim(-0.5, 3.0)
        ax1.set_aspect('equal')

        cube =  Polygon(cube_points, alpha=0.3, color='pink')
        ax1.add_patch(cube)

        matrix0 = r"$ y = \begin{bmatrix} a & c \\ b & d \end{bmatrix} x $"
        ax2.text(0.5, 0.8, matrix0, fontsize=30, ha='center', va='center', color='w')

        matrix1_1 = r"$ \begin{bmatrix} \_ \\ \_ \end{bmatrix} = \begin{bmatrix} a & c \\ b & d \end{bmatrix} \begin{bmatrix} 1 \\ 0 \end{bmatrix} $"
        matrix1_2 = r"$ \begin{bmatrix} a \\ \_ \end{bmatrix} = \begin{bmatrix} a & c \\ b & d \end{bmatrix} \begin{bmatrix} 1 \\ 0 \end{bmatrix} $"
        matrix1_3 = r"$ \begin{bmatrix} a \\ b \end{bmatrix} = \begin{bmatrix} a & c \\ b & d \end{bmatrix} \begin{bmatrix} 1 \\ 0 \end{bmatrix} $"

        mt1 = ax2.text(0.5, 0.5, matrix1_1, fontsize=30, ha='center', va='center', color='w')

        matrix2_1 = r"$ \begin{bmatrix} \_ \\ \_ \end{bmatrix} = \begin{bmatrix} a & c \\ b & d \end{bmatrix} \begin{bmatrix} 0 \\ 1 \end{bmatrix} $"
        matrix2_2 = r"$ \begin{bmatrix} c \\ \_ \end{bmatrix} = \begin{bmatrix} a & c \\ b & d \end{bmatrix} \begin{bmatrix} 0 \\ 1 \end{bmatrix} $"
        matrix2_3 = r"$ \begin{bmatrix} c \\ d \end{bmatrix} = \begin{bmatrix} a & c \\ b & d \end{bmatrix} \begin{bmatrix} 0 \\ 1 \end{bmatrix} $"

        mt2 = ax2.text(0.5, 0.2, matrix2_1, fontsize=30, ha='center', va='center', color='w')

        def on_reset():
            print('on_reset')
            mt1.set_text(matrix1_1)
            mt2.set_text(matrix2_1)

        collection = Collection(self.num_frames, ret_artists=[mt1, mt2], on_reset=on_reset)

        anim1_4 = Animation(0, 20)
        arrow1 = Arrow(ax1, cube_points[1], (0,0), width=0.07, head_width=0.2, head_length=0.2)
        arrow1.set_target_position(cube_points[1], np.array(data_points[0]) - np.array(cube_points[0]))
        anim1_4.append(arrow1)
        collection.append(anim1_4)

        # a <= (1,0)
        v10 = (['$1$', '$0$'], [(0.67, 0.53), (0.67, 0.44)])
        tg1 = TextGroup(ax2, *v10, fontsize=30)
        tg1.set_target_positions([(0.47, 0.53), (0.54, 0.53)])

        anim1 = Animation( 0,  5)
        anim1.append(tg1)
        collection.append(anim1)

        # a <= a
        tg2 = TextGroup(ax2, ['$a$'], [(0.47, 0.53)], fontsize=30)
        tg2.set_target_positions([(0.26, 0.53)])
        anim2 = Animation( 5, 10, on_stop=lambda: mt1.set_text(matrix1_2))
        anim2.append(tg2)
        collection.append(anim2)

        # b <= (1,0)
        tg3 = TextGroup(ax2, *v10, fontsize=30)
        tg3.set_target_positions([(0.47, 0.44), (0.54, 0.44)])

        anim3 = Animation(10, 15)
        anim3.append(tg3)
        collection.append(anim3)

        # b <= b
        tg4 = TextGroup(ax2, ['$b$'], [(0.47, 0.43)], fontsize=30)
        tg4.set_target_positions([(0.26, 0.43)])

        anim4 = Animation(15, 20, on_stop=lambda: mt1.set_text(matrix1_3))
        anim4.append(tg4)
        collection.append(anim4)

        # c <= (0,1)
        v01 = ['$0$', '$1$'], [(0.67, 0.23), (0.67, 0.13)]
        tg5 = TextGroup(ax2, *v01, fontsize=30)
        tg5.set_target_positions([(0.47, 0.23), (0.54, 0.23)])
        anim5 = Animation(20, 25)
        anim5.append(tg5)
        collection.append(anim5)

        # c <= c
        tg6 = TextGroup(ax2, ['$c$'], [(0.54, 0.23)], fontsize=30)
        tg6.set_target_positions([(0.26, 0.23)])
        anim6 = Animation(25, 30, on_stop=lambda: mt2.set_text(matrix2_2))
        anim6.append(tg6)
        collection.append(anim6)

        # d <= (0,1)
        tg7 = TextGroup(ax2, *v01, fontsize=30)
        tg7.set_target_positions([(0.47, 0.13), (0.54, 0.13)])
        anim7 = Animation(30, 35)
        anim7.append(tg7)
        collection.append(anim7)

        # d <= d
        tg8 = TextGroup(ax2, ['$d$'], [(0.54, 0.13)], fontsize=30)
        tg8.set_target_positions([(0.26, 0.13)])
        anim8 = Animation(35, 40, on_stop=lambda: mt2.set_text(matrix2_3))
        anim8.append(tg8)
        collection.append(anim8)

        anim9 = Animation(40, 45)
        collection.append(anim9)

        # print('index=', collection.index)
        para1 = Parallelogram(ax1, cube_points, alpha=0.3, color='cyan')
        para1.set_target_vertices([(0,0)] + data_points)
        anim_10 = Animation(45, 60)
        anim_10.append(para1)
        collection.append(anim_10)

        self.anim = collection.make_animation(self.fig)

    def save(self):
        import os
        import molass_legacy.KekLib.CustomMessageBox as MessageBox
        path = os.path.join(os.getcwd(), "anim.mp4")
        self.anim.save(path, writer="ffmpeg")
        MessageBox.showinfo("Save Notification", 'The movie has been saved to "%s"' % path, parent=self)



"""
    QuickAnalysis.ParammedModel.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import numpy as np

class ParammedModel:
    def __init__(self, model_params):
        self.model_params = model_params

    def append(self, model_spec):
        self.model_params.append(model_spec)

    def __call__(self, x):
        y = np.zeros(len(x))

        for model, params in self.model_params:
            y += model.eval(params, x=x)

        return y



"""
    Optimizer.ParamsIterator.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt

def iterator_test_from_dialog(caller):
    print("ParamsIterator.iterator_test_from_dialog")
    fv_list, xmax = caller.canvas.get_fv_array()
    fvvect = np.array([rec[1] for rec in fv_list])
    create_iterator(fvvect, debug=True)

def create_iterator(fvvect, debug=False):
    print("ParamsIterator.create_iterator: fvvect=", fvvect)
    last_fv = None
    iter_list = []
    for i in range(len(fvvect)-1):
        k = np.argmin(fvvect[0:i+1])
        if last_fv is None or fvvect[k] < last_fv:
            last_fv = fvvect[k]
            iter_list.append(k)

    if debug:
        x = np.arange(len(fvvect))
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("create_iterator: debug")
            ax.plot(x, fvvect)
            ax.plot(x[iter_list], fvvect[iter_list], "ro")
            fig.tight_layout()
            plt.show()
    
    return iter_list


"""
    SecTheory.T0UpperBound.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from bisect import bisect_right
import molass_legacy.KekLib.DebugPlot as plt
from GeometryUtils import rotated_argmin
from molass_legacy._MOLASS.SerialSettings import get_setting

SMALL_RATIO = 0.005
SMALL_ANGLE = -np.pi/24

def estimate_t0upper_bound(ecurve, debug=False):
    x = ecurve.x
    y = ecurve.y

    m = ecurve.peak_info[0][1]
    y_ = y[0:m]
    i = rotated_argmin(SMALL_ANGLE, y_)

    cy = np.cumsum(y)
    j = bisect_right(cy, cy[-1]*SMALL_RATIO)

    k = min(i, j)

    if debug:
        from PlotUtils import align_yaxis_np
        from DataUtils import get_in_folder

        in_folder = get_in_folder()

        with plt.Dp():
            fig, ax1 = plt.subplots()
            axt = ax1.twinx()
            axt.grid(False)
            fig.suptitle("T0Upper Bound Debug Plot for %s" % in_folder, fontsize=20)
            ax1.plot(x, y)
            ax1.plot(x[m], y[m], "o", color="green")
            ax1.plot(x[i], y[i], "o", color="red")
            ymin, ymax = ax1.get_ylim()
            ax1.set_ylim()
            ax1.plot(x[[k,k]], [ymin, ymax], color="yellow")
            axt.plot(x, cy, color="C1")
            axt.plot(x[j], cy[j], "o", color="cyan")
            align_yaxis_np(ax1, axt)
            fig.tight_layout()
            plt.show()

    return int(x[k])    # int(...) is meant to avoid str(...) to become like 'np.int64(...)' in a larger RAM environment



"""
    Models.Stochastic.DispersiveUtils.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Models.Stochastic.DispersivePdf import dispersive_monopore_pdf, DEFUALT_TIMESCALE

NUM_SDMCOL_PARAMS = 5
NUM_SDMCUV_PARAMS = 6

def compute_elution_curves(x, sdm_params, peak_rgs, timescale=DEFUALT_TIMESCALE, debug=False):
    me = 1.5
    mp = 1.5
    N_, K_, x0_, poresize_, N0_, tI_ = sdm_params[0:NUM_SDMCUV_PARAMS]
    # print("sdm_params[0:NUM_SDMCUV_PARAMS]=", sdm_params[0:NUM_SDMCUV_PARAMS])
    T_ = K_/N_
    # print("N_, T_, N0_, tI_=", N_, T_, N0_, tI_)
    scales_ = sdm_params[NUM_SDMCUV_PARAMS:]
    rhov = peak_rgs/poresize_
    rhov[rhov > 1] = 1
    cy_list = []
    t0_ = x0_ - tI_
    x_ = x - tI_
    if debug:
        print("x_[0]=", x_[0])
        moments_list = []
    for k, (rho, scale) in enumerate(zip(rhov, scales_)):
        ni_ = N_*(1 - rho)**me
        ti_ = T_*(1 - rho)**mp
        if debug:
            M1_ = x0_ + ni_*ti_
            M2_ = np.sqrt(2*ni_*ti_**2 + (t0_ + ni_*ti_)**2/N0_)
            moments_list.append([M1_, M2_])
            print([k], ni_, ti_, N0_, t0_)
        cy = scale*dispersive_monopore_pdf(x_, ni_, ti_, N0_, t0_, timescale=timescale)
        cy_list.append(cy)    
    ty = np.sum(cy_list, axis=0)

    if debug:
        print("compute_elution_curves: scales_", scales_)
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("compute_elution_curves")
            ax.plot(x, ty)
            for cy, (m1, s) in zip(cy_list, moments_list):
                ax.plot(x, cy, ":")
                ax.axvline(x=m1, color='yellow')
                ax.axvspan(m1-s, m1+s, color='yellow', alpha=0.3)
            ax.axvline(x=t0_, color='red')
            fig.tight_layout()
            plt.show()

    return cy_list, ty

def investigate_sdm_params(x, y, sdm_params, num_unreliables, peak_rgs, bounds):
    from matplotlib.widgets import Slider, Button

    # print("sdm_params=", sdm_params)
    print("peak_rgs=", peak_rgs)

    N_, K_, x0_, poresize_, N0_, tI_  = sdm_params[0:NUM_SDMCUV_PARAMS]
    print("K_=", K_)
    print("poresize_=", poresize_)
    print("N0_, tI_=", N0_, tI_)
    T_ = K_/N_
    me = 1.5
    mp = 1.5
    t0 = x0_ - tI_
    print("t0=", t0)
    slider_params = np.concatenate([sdm_params, [DEFUALT_TIMESCALE]])

    with plt.Dp():
        fig, ax = plt.subplots(figsize=(18, 5))
        ax.set_title("investigate_sdm_params")
        ax.plot(x, y)
        curves = []
        cy_list, ty = compute_elution_curves(x, slider_params[:-1], peak_rgs, timescale=slider_params[-1], debug=False)
        print("len(cy_list)=", len(cy_list))
        for cy, rg in zip(cy_list, peak_rgs):
            curve, = ax.plot(x, cy, ":", label="$R_g=$ %.3g" % rg)
            curves.append(curve)
        tcurve, = ax.plot(x, ty, ":", label="total", color="red")
        ax.legend()
        lines = []
        line = ax.axvline(x=tI_, color='gray')
        lines.append(line)
        line = ax.axvline(x=x0_, color='red')
        lines.append(line)
        line = ax.axvline(x=x0_ + K_, color='green')
        lines.append(line)
        fig.tight_layout()

        fig.subplots_adjust(right=0.7)

        slider_specs = [    (r"$N$", *bounds[0], slider_params[0]),
                            (r"$K$", *bounds[1], slider_params[1]),
                            (r"$x_0$", *bounds[2], slider_params[2]),
                            (r"poresize", *bounds[3], slider_params[3]),
                            (r"$N_0$", 1000, 80000, slider_params[4]),
                            (r"$t_I$", *bounds[4], slider_params[5]),
                            (r"$timescale$", 0.01, 1, slider_params[-1]),
                            ]

        def slider_update(k, val):
            # print([k], "slider_update", val)
            slider_params[k] = val
            N_, K_, x0_, poresize_, N0_, tI_ = slider_params[0:NUM_SDMCUV_PARAMS]
            timescale = slider_params[-1]
            cy_list, ty = compute_elution_curves(x, slider_params[:-1], peak_rgs, timescale=timescale)
            for curve, cy in zip(curves, cy_list):
                curve.set_data(x, cy)
            tcurve.set_data(x, ty)
            for line, pos in zip(lines, [tI_, x0_, x0_ + K_]):
                line.set_xdata(pos)
            fig.canvas.draw_idle()

        left = 0.75
        width = 0.2
        slider_axes = []
        sliders = []

        for k, (label, valmin, valmax, valinit) in enumerate(slider_specs, start=0):
            ax_ = fig.add_axes([left, 0.8 - 0.08*k, width, 0.03])
            slider  = Slider(ax_, label=label, valmin=valmin, valmax=valmax, valinit=valinit)
            slider.on_changed(lambda val, k_=k: slider_update(k_, val))
            slider_axes.append(ax_)
            sliders.append(slider)

        def reset(event):
            print("reset")
            for k, slider in enumerate(sliders):
                slider.reset()

        button_ax = fig.add_axes([0.85, 0.2, 0.12, 0.05])
        debug_btn = Button(button_ax, 'Reset', hovercolor='0.975')
        debug_btn.on_clicked(reset)

        ret = plt.show()
    
    return ret

def convert_to_sdm_guess_params(optimizer, p):
    xr_params, xr_baseparams, rg_params, (a, b), uv_params, uv_baseparams, (c, d), sdmcol_params = optimizer.split_params_simple(p)
    sdm_params = np.concatenate([sdmcol_params, xr_params])
    num_colparams = len(sdmcol_params)
    start_index = len(xr_params) + len(xr_baseparams)
    num_components = len(rg_params)
    bounds = list(optimizer.real_bounds[-num_colparams:]) + list(optimizer.real_bounds[start_index:start_index+num_components])
    return sdm_params, rg_params, bounds

def investigate_sdm_params_from_optimizer_params(optimizer, p):
    xr_curve = optimizer.xr_curve
    x = xr_curve.x
    y = xr_curve.y
    sdm_params, rg_params, bounds = convert_to_sdm_guess_params(optimizer, p)
    investigate_sdm_params(x, y, sdm_params, 0, rg_params, bounds)

def investigate_sdm_lrf_params_from_optimizer_params(optimizer, p):
    xr_curve = optimizer.xr_curve
    x = xr_curve.x
    y = xr_curve.y
    sdm_params, rg_params, bounds = convert_to_sdm_guess_params(optimizer, p)
    investigate_sdm_params(x, y, sdm_params, 0, rg_params, bounds)

def investigate_sdm_params_from_v2_result(caller):
    print("investigate_sdm_params_from_v2_result")
    # investigate_sdm_params(x, y, init_cuvparams, num_unreliables, temp_rgs, bounds)


"""
    DraggableCurves.py

    ref:
        https://stackoverflow.com/questions/63561034/arrows-with-draggable-end-in-matplotlib

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.backend_bases import MouseButton
from matplotlib.patches import ArrowStyle

class DraggableCurves:
    def __init__(self, ax, x_only=False, button_release_user_callback=None, debug=False):
        self.ax = ax
        self.fig = ax.figure
        self.x_only = x_only
        self.button_release_user_callback = button_release_user_callback
        self.debug = debug
        self.clear()

    def clear(self):
        self.picked_artist = None
        self.last_moveevent = None
        self.curve = None
        self.cursor = None
        self.moving = False
        self.button_release_callback(None)
        self.curve_list = []
        self.displ_list = []

    def add(self, x, y, *args, **kwargs):
        curve, = self.ax.plot(x, y, *args, **kwargs, picker=True)
        self.curve_list.append(curve)
        self.displ_list.append(np.zeros(2))

    def create_cursor(self, artist):
        x, y = artist.get_data()
        self.cursor, = self.ax.plot(x, y, ":", color="gray", lw=3)
        self.moving = False

    def remove_cursor(self):
        if self.debug:
            print("remove_cursor:", self.cursor)
        if self.cursor is None:
            # in the initial call
            pass
        else:
            self.cursor.remove()
        self.cursor = None
        self.moving = False

    def update_cursor(self, artist):
        if self.cursor is None:
            self.create_cursor(artist)
        else:
            if artist == self.picked_artist:
                # picking twice is regarded to be moving for the cursor to be able to be removed
                self.moving = True
            else:
                self.remove_cursor()
                self.create_cursor(artist)

    def update_displ_list(self):
        i = None
        for k, artist in enumerate(self.curve_list):
            if artist == self.picked_artist:
                i = k
                break
        assert i is not None
        event = self.last_moveevent
        target_point = np.array([event.xdata, event.ydata])
        self.displ_list[i] += target_point - self.picked_point
        self.last_displacement = self.displ_list[i]

    def get_displacements(self):
        return self.displ_list

    def get_last_displacement(self):
        return self.last_displacement

    def move(self, event, artist):
        if event is None or event.xdata is None or artist is None:
            return

        self.moving = True
        target_point = np.array([event.xdata, event.ydata])
        dx, dy = target_point - self.picked_point

        if self.x_only:
            dy = 0

        if self.debug:
            print("move: (%.2g, %.2g) => (%.2g, %.2g)" % (*self.picked_point, *target_point))
            print("move: %.2g, %.2g" % (dx, dy))

        x, y = self.picked_artist.get_data()
        artist.set_data(x+dx, y+dy)
        self.last_moveevent = event

    def pick_callback(self, event):
        'called when an element is picked'
        if self.debug:
            print("pick_callback")
        if event.mouseevent.button == MouseButton.LEFT:
            self.picked_artist = event.artist
            self.update_cursor(self.picked_artist)

            mouseevent = event.mouseevent
            self.picked_point = np.array([mouseevent.xdata, mouseevent.ydata])
            self.fig.canvas.draw_idle()

    def button_release_callback(self, event):
        'called when a mouse button is released'
        if self.debug:
            print("button_release_callback")
        if event is None or event.button == MouseButton.LEFT:
            if self.moving:
                self.move(self.last_moveevent, self.picked_artist)
                self.update_displ_list()
                if self.button_release_user_callback is not None:
                    self.button_release_user_callback(self)
                self.picked_artist = None
                self.picked_point = None
                self.last_moveevent = None
                self.remove_cursor()
                self.fig.canvas.draw_idle()

    def motion_notify_callback(self, event):
        'called when the mouse moves'
        if self.picked_artist is not None and event.inaxes is not None and event.button == MouseButton.LEFT:
            if self.debug:
                print("motion_notify_callback: move")
            self.move(event, self.cursor)
            self.fig.canvas.draw_idle()

def demo():
    fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(10,4))
    ax1.set_title("DraggableCurves Demo")

    x = np.linspace(-np.pi, +np.pi, 100)

    curves = DraggableCurves(ax1, x_only=True, debug=False)
    curves.add(x, np.sin(x))
    curves.add(x, np.sin(2*x))

    ax1.set_xlim(-np.pi, +np.pi)
    ax1.set_ylim(-2, 2)

    fig.canvas.mpl_connect('button_release_event', curves.button_release_callback)
    fig.canvas.mpl_connect('pick_event', curves.pick_callback)
    fig.canvas.mpl_connect('motion_notify_event', curves.motion_notify_callback)

    plt.show()

if __name__ == '__main__':
    demo()



"""
    ExcelTeller.py

    Copyright (c) 2019-2024, SAXS Team, KEK-PF
"""
from molass_legacy._MOLASS.Processes import register_process, remove_process

class TellerInfo:
    def __init__(self, code, **kwargs):
        self.code = code
        self.__dict__.update(kwargs)

    def __repr__(self):
        return 'TellerInfo<%s>' % self.code

class ExcelTeller:
    def __init__(self, log_folder=None):
        import logging
        self.logger = logging.getLogger(__name__)
        self.log_folder = log_folder
        self.invoke_excel_manager()

    def invoke_excel_manager(self):
        import multiprocessing as mp
        from .ExcelManager import ExcelManager
        self.logger.info('invoke_excel_manager')
        self.excel_manager = ExcelManager()
        self.job_queue = mp.Queue()
        self.manager_process = mp.Process(target=self.excel_manager.manage_loop, args=(self.job_queue, self.log_folder))
        register_process(self.manager_process)
        self.manager_process.start()

    def get_pid(self):
        return self.manager_process.pid

    def stop(self):
        self.job_queue.put(TellerInfo('stop'))
        remove_process(self.manager_process)
        self.manager_process = None

    def tell(self, code, **kwargs):
        self.job_queue.put(TellerInfo(code, **kwargs))



"""
    Models.Stochastic.PsdIllust.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
import seaborn
seaborn.set_theme()

def draw_psd_illust():
    from molass_legacy.Models.Stochastic.LognormalPoreFunc import distr_func
    from molass_legacy.Models.Stochastic.LognormalUtils import compute_mu_sigma

    pv = np.linspace(0, 400, 100)
    fig, axes = plt.subplots(2, 2, figsize=(8, 6))

    mode, stdev = 200, 20
    mu, sigma = compute_mu_sigma(mode, stdev)
    ax = axes[1,1]
    ax.set_title("Lognormalpore")
    ax.fill_between(pv, np.zeros(len(pv)), distr_func(pv, mu, sigma))
    xmin, xmax = ax.get_xlim()

    ax = axes[0,0]
    ax.set_title("Monopore")
    ax.bar(200, 1, width=20)
    ax.set_xlim(xmin, xmax)

    ax = axes[0,1]
    ax.set_title("Dipore")
    ax.bar([70, 200], [0.3, 0.7], width=20)
    ax.set_xlim(xmin, xmax)

    ax = axes[1,0]
    ax.set_title("Tripore")
    ax.bar([70, 200, 300], [0.2, 0.5, 0.3], width=20)
    ax.set_xlim(xmin, xmax)

    for ax in axes.flatten():
        ax.set_xlabel(r"Poresize ($\AA$)")
        ax.set_ylabel("Density")

    fig.tight_layout()
    plt.show()

if __name__ == "__main__":
    import sys
    import os
    this_dir = os.path.dirname( os.path.abspath( __file__ ) )
    sys.path.append( this_dir + '/../..' )
    draw_psd_illust()


# -*- coding: utf-8 -*-
"""

    ファイル名：   TestDataGenerator.py

    処理内容：

        テスト用の模擬データを生成する。

"""
from __future__ import division, print_function, unicode_literals

import sys
import os
import shutil
import re
from nose.tools import eq_

import PilatusCounter
from PilatusUtils import get_data_info

import glob
import shutil

num_part_re     = re.compile( '^[^\d]+(\d+)' )
image_file_re   = re.compile( '^([^_]+)_' )
counter_file_re = re.compile( '^PilatusCounter_([^_]+)_' )

counter_id      = 'None'
pilatus_counter = None
adj_folder      = None
syn_folder      = None

class TestDataGenerator:
    def __init__( self, in_folder ):
        self.in_folder = in_folder

        curdir = os.getcwd()

        os.chdir( in_folder )

        logfiles = glob.glob( 'measurement*.log' )
        bakfiles = glob.glob( '*.log.bak' )

        assert( len( logfiles ) == 1 )

        self.logfile = logfiles[0]

        if len( bakfiles ) == 0:
            bakfile = logfiles[0].replace( '.log', '.log.bak' )
            shutil.copy2( logfiles[0], bakfile )

        os.chdir( curdir )

    def restore_from_bak( self ):
        curdir = os.getcwd()

        os.chdir( self.in_folder )

        bakfiles = glob.glob( '*.log.bak' )
        assert( len( bakfiles ) == 1 )
        logfile = bakfiles[0].replace( '.log.bak', '.log' )
        shutil.copy2( bakfiles[0], logfile )

        data_array, text_dict = get_data_info( self.in_folder, adj_folder, syn_folder, pilatus_counter, counter_id, for_test_data=True )

        remove_count = 0
        for file in glob.glob( '*.tif' ):
            # print( file )
            m = image_file_re.match( file )
            if m:
                sample_id = m.group( 1 )
                if not sample_id in text_dict:
                    remove_count += 1
                    os.remove( file )

        for file in glob.glob( 'PilatusCounter_*.txt' ):
            # print( file )
            m = counter_file_re.match( file )
            if m:
                sample_id = m.group( 1 )
                if not sample_id in text_dict:
                    remove_count += 1
                    os.remove( file )

        if remove_count > 0:
            print( 'restore_from_bak: Rmoved %d files.' % ( remove_count ) )

        os.chdir( curdir )

    def get_data_array( self ):

        data_array, text_dict = get_data_info( self.in_folder, adj_folder, syn_folder, pilatus_counter, counter_id, for_test_data=True )
        self.data_array = data_array
        self.text_dict  = text_dict
 
        return data_array

    def incremented_id( self, name ):
        m = num_part_re.match( name )
        if m:
            oldnum = m.group( 1 )
            len_ = len( oldnum )
            fmt = '%0{0}d'.format( len_ )
            newnum = fmt % ( int( oldnum ) + 1 )
            # print( oldnum, '=>', newnum  )
            return name.replace( oldnum, newnum )
        else:
            assert( False )

    def add_mockcopy( self, num_copies=1 ):

        curdir = os.getcwd()
        os.chdir( self.in_folder )

        for i in range( num_copies ):
            data_array = self.get_data_array()
            oldname = data_array[-1][0]
            newname = self.incremented_id( oldname )

            tiffiles = glob.glob( oldname + '*.tif' )
            for oldtif in tiffiles:
                newtif = oldtif.replace( oldname, newname )
                shutil.copy2( oldtif, newtif )

            counterfiles = glob.glob( 'PilatusCounter_' + oldname + '*.txt' )
            assert( len( counterfiles ) == 1 )
            new_file = counterfiles[0].replace( oldname, newname )
            fh = open( counterfiles[0] )
            content = fh.read()
            fh.close()
            fh = open( new_file, 'w' )
            fh.write( content.replace( oldname, newname ) )
            fh.close()

            logtext = self.text_dict[ oldname ]
            # print( 'logtext=', logtext )
            fh = open( self.logfile, 'a' )
            fh.write( "\n" )
            fh.write( logtext.replace( oldname, newname ) )
            fh.close()
            print( 'Mock %s has been created.' % (newname) )

        os.chdir( curdir )

if __name__ == '__main__':
    mdg = MockDataGenerator()
    mdg.generate( 1 )
    


"""
    Optimizer.Strategies.DefaultStrategy.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

class DefaultStrategy:
    def __init__(self):
        pass

    def trust_initial_baseline(self):
        return True

    def baseline_first(self):
        return True

    def is_strategic(self, n):
        return False


"""
    SecTheory.CavazziniProof.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
from datetime import datetime
import molass_legacy.KekLib.DebugPlot as plt
from .SecCF import two_site_phi
from .SecPDF import FftInvPdf

class TwoSiteSpikeModel:
    def __init__(self, m):
        self.m = m
        self.pdf = FftInvPdf(two_site_phi)

    def __call__(self, x, n, p1, tratio):
        t1 = self.m/n
        p2 = 1 - p1
        t2 = tratio*t1
        return self.pdf(x, n, p1, p2, t1, t2)

class TwoSiteModel:
    def __init__(self):
        self.pdf = FftInvPdf(two_site_phi)

    def fit(self, x, y, init_params=None):
        h_ = 100
        n_ = len(x)//2
        p1_ = 0.99
        t1_ = 1
        t2_ = 10*t1_
        def obj_func(p):
            h, n, p1, t1, t2 = p
            neg_penalty = min(0, p1)**2 + (max(1, p1) - 1)**2 + min(0, t1)**2 + min(0, t2)**2
            return np.sum((h*self.pdf(x, n, p1, 1 - p1, t1, t2) - y)**2) + neg_penalty*1e8

        # init_params = [2.01175632e+02, 6.53814668e+02, 9.24384125e-01, 1.49033153e-01,  1.21771676e+01]
        # init_params = [100, 500, 0.99, 0.1,  1]
        # init_params = [201.9764699, 530.36354375,  0.96139286, 0.61745049, 18.6660736]
        if init_params is None:
            init_params = [200, 500, 0.99, 0.5, 5]

        # ret = basinhopping(obj_func, (h_, n_, p1_, t1_, t2_))
        ret = basinhopping(obj_func, init_params)
        # ret = basinhopping(obj_func, init_params)
        return ret

    def __call__(self, x, h, n, p1, t1, t2):
        return h*self.pdf(x, n, p1, 1 - p1, t1, t2)

class TwoSiteAlphaModel:
    def __init__(self):
        self.pdf = FftInvPdf(two_site_phi)

    def fit(self, x, y, init_params=None):
        def obj_func(p):
            h, n, p1, alpha = p
            neg_penalty = min(0, p1)**2 + (max(1, p1) - 1)**2 + min(0, alpha)**2
            return np.sum((h*self.pdf(x, n, p1, 1 - p1, 1, alpha) - y)**2) + neg_penalty*1e8

        # init_params = [2.01175632e+02, 6.53814668e+02, 9.24384125e-01, 1.49033153e-01,  1.21771676e+01]
        # init_params = [100, 500, 0.99, 0.1,  1]
        # init_params = [201.9764699, 530.36354375,  0.96139286, 0.61745049, 18.6660736]
        if init_params is None:
            init_params = [211, 574, 0.995, 65.6]

        # ret = basinhopping(obj_func, (h_, n_, p1_, t1_, t2_))
        ret = basinhopping(obj_func, init_params)
        # ret = basinhopping(obj_func, init_params)
        return ret

    def __call__(self, x, h, n, p1, alpha):
        return h*self.pdf(x, n, p1, 1 - p1, 1, alpha)


def demo0():
    t = np.arange(300)

    two_site_pdf = FftInvPdf(two_site_phi)

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.set_title("Two Site Model", fontsize=20)
        # n, p1, p2, t1, t2 = 100, 0.99, 0.01, 0.1, 1
        p1, p2 = 0.99, 0.01
        t1 = 0.1
        t2 = 1
        for n in [100, 200, 400, 800, 1000, 2000]:
            # tR = n*(p1*t1 + p2*t2)
            ax.plot(t, two_site_pdf(t, n, p1, p2, t1, t2), label="n=%d" % n)
        # xmin, xmax = ax.get_xlim()
        # ax.set_xlim(0, 3)
        ax.legend()
        fig.tight_layout()
        plt.show()

def demo1():
    t = np.arange(300)

    model = TwoSiteSpikeModel(150)

    with plt.Dp():
        fig, ax = plt.subplots()
        ax.set_title("Two Site Model", fontsize=10)
        # n, p1, p2, t1, t2 = 100, 0.99, 0.01, 0.1, 1
        p1 = 0.99
        p2 = 1 - p1
        for n in [100, 200, 400, 800, 1000, 2000]:
            ax.plot(t, model(t, n, p1, 10), label="n=%d" % n)
        ax.legend()
        fig.tight_layout()
        plt.show()

def real_demo(in_folder, trimming=True, correction=True):
    from scipy.optimize import curve_fit
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Baseline.BaselineUtils import get_corrected_sd_impl
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from molass_legacy.Elution.CurveUtils import simple_plot
    from molass_legacy.Peaks.ElutionModels import egh
    from LPM import get_corrected
    from DataUtils import get_in_folder

    sp = StandardProcedure()
    sd = sp.load_old_way(in_folder)

    pre_recog = PreliminaryRecognition(sd)
    if trimming:
        sd_ = sd._get_analysis_copy_impl(pre_recog)
    else:
        sd_ = sd.get_copy()

    if correction:
        v2_copy = get_corrected_sd_impl(sd_, sd, pre_recog)
    else:
        v2_copy = sd_

    D, _, wv, ecurve = v2_copy.get_uv_data_separate_ly()

    ex = ecurve.x[10:-10]
    ey = get_corrected(ecurve.y[10:-10])

    p_init = ecurve.get_emg_peaks()[ecurve.primary_peak_no].get_params()
    popt1, pcov1 = curve_fit(egh, ex, ey, p_init)


    if False:
        model = MonoPoreModel()
        # init_params = [10, 150, 0.99, 0.1, 1]
        # init_params = [21.09085630844535, 57.05478375607099, 0.8134837337129833, 2.8055854804291713, 2.8055891557224237]
        # init_params = [21.09085438503959, 57.05479161605151, 0.4742816953836163, 2.8055865820081425, 2.805584983307841]
        # init_params = [21.0908540732968, 57.05479955016449, 0.8037212374299743, 2.8055851256770192, 2.805586333848221]
        # init_params = [21.0908535, 57.05480977, 2.80558488]
        init_params = [21.09085432546422, 57.05479510560458, 2.8055856104192967]
    elif in_folder.find("OA") > 0:
        model = TwoSiteModel()
        # init_params = [10, 150, 0.99, 0.1, 1]
        # init_params = [21.09085630844535, 57.05478375607099, 0.8134837337129833, 2.8055854804291713, 2.8055891557224237]
        # init_params = [21.09085438503959, 57.05479161605151, 0.4742816953836163, 2.8055865820081425, 2.805584983307841]
        init_params = [21.0908540732968, 57.05479955016449, 0.8037212374299743, 2.8055851256770192, 2.805586333848221]
    else:
        model = TwoSiteModel()
        # init_params = [200, 500, 0.99, 0.5, 5]
        # init_params = [212.06604355, 496.60195744, 0.99516111, 1.15354905,  67.06323863]
        # params= [211.98227569 497.36669375   0.99517607   1.15150084  67.23597666]
        # params= [212.03260787 500.283879     0.99520864   1.14479576  67.35694569]
        # params= [211.76525286 501.38064212   0.99517714   1.14140277  66.72618185]
        # params= [211.93850687 503.1646396    0.99521843   1.1377971   67.13311391]
        # init_params= [211.82262009, 506.42648153, 0.99522858, 1.12996051, 66.8657622]
        # init_params= [211.76262156348102, 522.2824122494156, 0.9953431949768717, 1.0945862185945257, 66.67018690354158]
        # init_params= [211.51931787620344, 542.0343851514614, 0.9954524856411976, 1.0529220645070299, 66.03641538852862]
        # init_params= [211.5125903865166, 558.5731370008768, 0.9955732039827447, 1.0211441250646398, 66.03620025776257]
        # init_params= [211.32184877429745, 566.583019797747, 0.9955967593428952, 1.0057396934143517, 65.56377332774754]
        # init_params = [211.53926986961773, 580.567581537651, 0.9957255508661492, 0.9817207281960352, 66.0832372341951]
        # init_params = [211.23380012691356, 589.007836848129, 0.9957328806835914, 0.9663319382212798, 65.30489077398083]
        # init_params = [211.1730691340855, 605.1641584047355, 0.995826281975228, 0.9398468190500091, 65.15391941190406]
        # init_params = [211.01281199499408, 614.712295404362, 0.9958630063530421, 0.9245231583976347, 64.78366069670133]
        # init_params = [210.80233635855276, 664.0508285684691, 0.996111858418305, 0.8539738842429706, 64.25952300122583]
        # init_params = [210.519377628442, 707.2090960646246, 0.9962884635747533, 0.800127310961128, 63.541234776791946]
        # init_params = [210.29024437581631, 728.897932727702, 0.9963635408056852, 0.7754107393393052, 63.047159820670586]
        # bad init_params = [210.28559177527205, 728.4806403663978, 0.9963682614399137, 0.7760005009963438, 63.17116889910384]
        # init_params = [210.2432129644718, 736.1089550848279, 0.9964039886713507, 0.7678615697396018, 63.17276056221165]
        # init_params = [210.31577022660957, 736.8047259033731, 0.9964016280718834, 0.7669956193079287, 63.09645579362846]
        # init_params = [210.31861692248353, 738.8852507261381, 0.9964109159558705, 0.7648006217330974, 63.09806054328252]
        # bad init_params = [210.340091473509, 739.4386070044351, 0.996413292144056, 0.7642248442437305, 63.10330054692467]
        # init_params = [210.28932436757205, 739.3193830268104, 0.9964112448124383, 0.7643298019955326, 63.05725223453272]
        # bad init_params = [210.3086178076916, 742.9637121509476, 0.9964273491334341, 0.7604953626068545, 63.07460696639221]
        # init_params = [210.34811156158634, 743.3092486833381, 0.9964326679615353, 0.7602056546113496, 63.1493213914855]
        # init_params = [210.2747478064457, 747.6041965051638, 0.9964439548938019, 0.7556281120973185, 62.99773805323856]
        init_params = [210.2452597184559, 749.4992507845662, 0.9964492328705874, 0.7536282577914788, 62.939887204560435]

    fh = open("params.csv", "w")

    ok_params = []
    min_value = None

    for k in range(10000):
        ret = model.fit(ex, ey, init_params=init_params)
        params = ret.x
        print("ret.fun=", ret.fun)

        if np.isnan(ret.fun):
            print("******************************** bad params =", params)
            init_params = ok_params[-1]
            continue

        if k % 1000 == 0:
            print(datetime.now().strftime("%Y-%m-%d %H:%M:%S"), [k], "---------------------------------------------")

        # if k % 100 == 0:
        if k == 0:
            with plt.Dp():
                in_folder = get_in_folder(in_folder)
                fig, ax = plt.subplots()
                ax.set_title("Two-site Stochastic Model Fitting for %s" % in_folder, fontsize=20)
                # simple_plot(ax, ecurve, legend=False)
                ax.plot(ex, ey, label="data")
                ax.plot(ex, egh(ex, *popt1), ":", lw=3, label="egh")
                ax.plot(ex, model(ex, *params), ":", lw=3, label="stochastic model")
                ax.legend(fontsize=16)
                fig.tight_layout()
                plt.show()

        if min_value is None or ret.fun < min_value:
            print([k], "------------------ ret.fun=", ret.fun)
            print("params=", params)
            fh.write("[%d] fv=%s\n" % (k, str(ret.fun)))
            fh.write("params=" + ",".join([str(p) for p in params]) + "\n")
            fh.flush()
            init_params = params
            min_value = ret.fun
            ok_params.append(params)

    fh.close()

def real_demo1(in_folder, trimming=True, correction=True):
    from scipy.optimize import curve_fit
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Baseline.BaselineUtils import get_corrected_sd_impl
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from molass_legacy.Elution.CurveUtils import simple_plot
    from molass_legacy.Peaks.ElutionModels import egh
    from LPM import get_corrected
    from DataUtils import get_in_folder

    sp = StandardProcedure()
    sd = sp.load_old_way(in_folder)

    pre_recog = PreliminaryRecognition(sd)
    if trimming:
        sd_ = sd._get_analysis_copy_impl(pre_recog)
    else:
        sd_ = sd.get_copy()

    if correction:
        v2_copy = get_corrected_sd_impl(sd_, sd, pre_recog)
    else:
        v2_copy = sd_

    D, _, wv, ecurve = v2_copy.get_uv_data_separate_ly()

    ex = ecurve.x[10:-10]
    ey = get_corrected(ecurve.y[10:-10])

    p_init = ecurve.get_emg_peaks()[ecurve.primary_peak_no].get_params()
    popt1, pcov1 = curve_fit(egh, ex, ey, p_init)

    model = TwoSiteAlphaModel()

    fh = open("params.csv", "w")

    init_params = None
    min_value = None

    for k in range(1000):
        ret = model.fit(ex, ey, init_params=init_params)
        params = ret.x
        # print("params=", params)

        if k % 100 == 0:
            with plt.Dp():
                in_folder = get_in_folder(in_folder)
                fig, ax = plt.subplots()
                ax.set_title("Two-site Stochastic Model Fitting for %s" % in_folder, fontsize=20)
                # simple_plot(ax, ecurve, legend=False)
                ax.plot(ex, ey, label="data")
                ax.plot(ex, egh(ex, *popt1), ":", lw=3, label="egh")
                ax.plot(ex, model(ex, *params), ":", lw=3, label="stochastic model")
                ax.legend(fontsize=16)
                fig.tight_layout()
                plt.show()

        if min_value is None or ret.fun < min_value:
            print([k], "------------------ ret.fun=", ret.fun)
            print("params=", params)
            fh.write("[%d] fv=%s\n" % (k, str(ret.fun)))
            fh.write("params=" + ",".join([str(p) for p in params]) + "\n")
            fh.flush()
            init_params = params
            min_value = ret.fun

    fh.close()

def fitting_demo_impl(in_folder, stochastic_model, init_params):
    from scipy.optimize import curve_fit
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Baseline.BaselineUtils import get_corrected_sd_impl
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from molass_legacy.Peaks.ElutionModels import egh
    from LPM import get_corrected
    from DataUtils import get_in_folder
    from SecSaxs.DataTreatment import DataTreatment
    from .StochasticSolver import CfDomain

    sp = StandardProcedure()
    sd = sp.load_old_way(in_folder)

    pre_recog = PreliminaryRecognition(sd)
    treat = DataTreatment(route="v2", trimming=2, correction=1)
    sd_copy = treat.get_treated_sd(sd, pre_recog)

    D, E, qv, ecurve = sd_copy.get_xr_data_separate_ly()

    fig, ax = plt.subplots()

    x = ecurve.x
    y = ecurve.y

    solver = CfDomain(x, y)
    spline = solver.fit(stochastic_model, init_params=init_params)

    ax.plot(x, y, label="data")
    ax.plot(x, spline(x), label="model")

    ax.legend()

    fig.tight_layout()
    plt.show()

def fitting_demo_single_site(in_folder, init_params):
    from .Cavazzini1999 import single_site_cf
    fitting_demo_impl(in_folder, single_site_cf, init_params)

def fitting_demo_two_site(in_folder, init_params):
    from .Cavazzini1999 import two_site_cf
    fitting_demo_impl(in_folder, two_site_cf, init_params)

def fitting_demo_three_site(in_folder, init_params):
    from .Cavazzini1999 import three_site_cf
    fitting_demo_impl(in_folder, three_site_cf, init_params)

def fitting_demo_uniform_adsorption_energy(in_folder, init_params):
    from .Cavazzini1999 import uniform_adsorption_energy
    fitting_demo_impl(in_folder, uniform_adsorption_energy, init_params)

def fitting_demo_uniform_mean_sojourn_time(in_folder, init_params):
    from .Cavazzini1999 import uniform_mean_sojourn_time
    fitting_demo_impl(in_folder, uniform_mean_sojourn_time, init_params)



"""
    Theory.PdbCrysolRoute.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import UnivariateSpline
from glob import glob
from molass_legacy.KekLib.BasicUtils import Struct
from DataUtils import get_in_folder
from CrysolUtils import np_loadtxt_crysol
from molass_legacy.ATSAS.Crysol import get_info_from_crysol_log, get_info_from_crysol_fit_log
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkCustomWidgets import FileEntry, FolderEntry
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.DENSS.DenssUtils import fit_data
from molass_legacy._MOLASS.Version import is_developing_version
from molass_legacy.Saxs.SaxsCurveUtils import percentile_normalize

drive = __file__[0:2]
crysol_folder = drive + r"\PyTools\Data\CRYSOL"
crysol_int_file = os.path.join(crysol_folder, "1wbh-ba1.int")

def demo(dialog):
    print("PdbCrysolRoute demo")

TEST_FOLDERS = [
    drive + r"\TODO\20231031\FIND\crysol_folders\2eph-ba1",
    # drive + r"\TODO\20231031\FIND\crysol_folders\6ald-ba1",
    # drive + r"\TODO\20231031\FIND\crysol_folders\6v20-ba1",
    drive + r"\TODO\20231031\FIND\crysol_folders\5f4x-ba1",
    ]

class CrysolResultsEntry(Dialog):
    def __init__(self, parent, num_files):
        self.num_files = num_files
        self.applied = False
        Dialog.__init__(self, parent, "Crysol Results Entry", visible=False, location="lower right")

    def show(self):
        self._show()

    def body(self, body_frame):

        entry_frame = Tk.Frame(body_frame)
        entry_frame.pack()

        row = 0

        text_vars = []
        for i in range(self.num_files):
            label = Tk.Label(entry_frame, text="CRYSOL Output Folder %s:" % str([i+1]))
            label.grid(row=row+i, column=0, padx=5, sticky=Tk.E)
            var = Tk.StringVar()
            if is_developing_version():
                var.set(TEST_FOLDERS[i])
            text_vars.append(var)
            entry = FolderEntry(entry_frame, textvariable=var, width=60)
            entry.grid(row=row+i, column=1, padx=5, sticky=Tk.W)

        self.text_vars = text_vars

    def get_file_paths(self):
        ret_paths = []
        for var in self.text_vars:
            ret_paths.append(var.get())
        return ret_paths

    def validate(self):
        num_paths = 0
        for var in self.text_vars:
            val = var.get()
            if val is None or val == "":
                pass
            else:
                num_paths += 1
        if num_paths == 0:
            import molass_legacy.KekLib.CustomMessageBox as MessageBox
            MessageBox.showerror('No entry',
                            "You should specify a file for at least one component.\n",
                            parent=self)
            return 0
        else:
            return 1

    def apply(self):
        self.applied = True

class PdbCrysolRouteDialog(Dialog):
    def __init__(self, parent, view_dialog, crysol_folders):
        self.view_dialog = view_dialog
        self.crysol_folders = crysol_folders
        Dialog.__init__(self, parent, "Pdb Crysol Route", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        to_draw_components = []
        to_draw_folders = []
        for k, path in enumerate(self.crysol_folders):
            if path is None or path == "":
                pass
            else:
                to_draw_components.append(k)
                to_draw_folders.append(path)

        num_components = len(to_draw_components)

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        fig, axes = plt.subplots(nrows=num_components, ncols=3, figsize=(18,4*num_components))
        self.draw_curves(fig, axes, to_draw_components, to_draw_folders)
        self.fig = fig

        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas.draw()

        tframe = Tk.Frame(body_frame)
        tframe.pack(side=Tk.LEFT, padx=20)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe )
        self.toolbar.update()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=50, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def get_info(self, folder):
        int_file = None
        short_name_len = None
        Rg = None
        chi_square = None
        for node in os.listdir(folder):
            if short_name_len is None:
                short_name_len = len(node)
            path = os.path.join(folder, node)
            ext = node[-4:]
            if ext == ".int":
                int_file = path
            elif ext == ".log":
                if len(node) == short_name_len:
                    info = get_info_from_crysol_log(path)
                    Rg = info.Rg
                else:
                    info = get_info_from_crysol_fit_log(path)
                    Chi_square = info.Chi_square
        return Struct(int_file=int_file, Rg=Rg, Chi_square=Chi_square)

    def draw_curves(self, fig, axes, to_draw_components, to_draw_folders):

        to_draw_infos = [self.get_info(f)  for f in to_draw_folders]

        in_folder = get_in_folder()

        view_dialog = self.view_dialog

        lrf_info = view_dialog.lrf_info
        x = lrf_info.x
        y = lrf_info.y
        Pxr, Cxr, Puv, Cuv, mapped_UvD = lrf_info.matrices

        zero_base = np.zeros(len(x))

        optimizer = view_dialog.optimizer
        qv = optimizer.qvector
        xrD = optimizer.xrD
        xrE = optimizer.xrE
        xr_i = optimizer.xr_index

        cy_list = []
        peak_positions = []
        my_curves = []
        py_curves = []
        py_ift_list = []
        for k, cy in enumerate(Cxr):
            m = np.argmax(cy)
            peak_positions.append(m)
            my = xrD[:,m]
            my_curves.append(my)
            scale = Pxr[xr_i,k]
            cy_list.append(scale*cy)
            if k in to_draw_components:
                py = Pxr[:,k]
                py_curves.append(py)
                sasrec = fit_data(qv,py, xrE[:,m], return_sasrec=True)
                py_ift_list.append((sasrec.qc, sasrec.Ic, sasrec.Icerr, sasrec.rg))
        ty = np.sum(cy_list, axis=0)

        fig.suptitle("Comparison to PDB-generated Curves on %s" % (in_folder), fontsize=20)

        if len(axes) == 3:
            ax1, ax2, ax3 = axes
            axes_ = [axes]
        else:
            ax1, ax2, ax3 = axes[0,:]
            axes_ = axes
        ax1.set_title("Component Elution", fontsize=16)
        ax2.set_title("Peak Top, LRF and IFT", fontsize=16)
        ax3.set_title("IFT and CRYSOL", fontsize=16)

        for k, (ax1, ax2, ax3) in enumerate(axes_):
            info = to_draw_infos[k]
            path = info.int_file
            data = np_loadtxt_crysol(path)[0]
            dk = to_draw_components[k]

            ax1.plot(x, y)
            for j, cy_ in enumerate(cy_list):
                if j == dk:
                    ax1.fill_between(x, zero_base, cy_, fc="cyan", alpha=0.2)
                ax1.plot(x, cy_, ":")
            ax1.plot(x, ty, ":", color="red")

            ax1.set_xlabel("Eno")
            ax1.set_ylabel("Intensity")

            for ax in (ax2, ax3):
                ax.set_yscale('log')
                ax.set_xlabel("Q")
                ax.set_ylabel("$Log_{10}(Intensity)$")

            ax2.plot(qv, percentile_normalize(my_curves[k]), label="peak top")
            ax2.plot(qv, percentile_normalize(py_curves[k]), label="LRF")
            qc, Ic, Icerr, rg = py_ift_list[k]
            nIc = percentile_normalize(Ic)

            for ax in [ax2, ax3]:
                ax.plot(qc, nIc, label="LRF IFT; $R_g$=%.1f" % rg, color="C2")

            qv_ = data[:,0]
            py_ = percentile_normalize(data[:,1])
            file = path.replace('/', '\\').split('\\')[-1].replace('.int', '')
            file_ = file[:4].upper() + file[4:].lower()
            ax3.plot(qv_, py_, label=r"%s; $R_g$=%.1f, $\chi^2$=%.3g" % (file_, info.Rg, info.Chi_square), color="C3")

            for ax in [ax2, ax3]:
                ax.legend()

            crysol_spline = UnivariateSpline(qv_, py_, s=0, ext=3)
            residual = (nIc - crysol_spline(qc))/Icerr
            axt = ax3.twinx()
            axt.grid(False)
            axt.bar(qc, residual, width=0.002, color='purple', alpha=0.1, label='residual (diff/error)')
            ymin, ymax = axt.get_ylim()
            maxabs = max(abs(ymin), abs(ymax))
            axt.set_ylim(-maxabs, maxabs)
            axt.legend(bbox_to_anchor=(1, 0.4), loc='upper right')

        fig.tight_layout()


def compare_to_pdb_impl(view_dialog):
    print("PdbCrysolRoute demo")
    entry = CrysolResultsEntry(view_dialog, 2)
    entry.show()
    if not entry.applied:
        return

    crysol_folders = entry.get_file_paths()

    dialog = PdbCrysolRouteDialog(view_dialog.parent, view_dialog, crysol_folders)
    dialog.show()



"""
    ErrorCorrectionDemo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
import numpy as np
from scipy.optimize import minimize
from bisect import bisect_right
from scipy.stats import linregress
from scipy.interpolate import UnivariateSpline
from matplotlib.patches import Rectangle
import molass_legacy.KekLib.DebugPlot as plt
from Theory.SolidSphere import phi
from time import time

# SAVE_FOLDER = r"D:\TODO\20230710\temp"
SAVE_FOLDER = None

def save_aq(file, qv, P, Pe):
    if SAVE_FOLDER is not None:
        path = os.path.join(SAVE_FOLDER, file)
        np.savetxt(path, np.array([qv, P[:,0], Pe[:,0]]).T)

def demo_impl(sd, f, t):
    from DataUtils import get_in_folder
    from molass_legacy.DENSS.DenssUtils import fit_data
    from .NaiiveLrfSolver import NaiiveLrfSolver
    from .BoundedLrfSolver import BoundedLrfSolver
    from importlib import reload
    import BoundedLRF.ErrorCorrection
    reload(BoundedLRF.ErrorCorrection)
    from .ErrorCorrection import compute_corrected_error

    print("demo_impl")

    D, E, qv, xr_curve = sd.get_xr_data_separate_ly()
    x = xr_curve.x
    y = xr_curve.y
    data_title = get_in_folder()

    range_ = slice(f, t+1)

    D_, E_ = D[:,range_], E[:,range_]

    i = bisect_right(qv, 0.02)
    cy = D_[i,:]
    j = np.argmax(cy)

    print("len(cy)=", len(cy), "j=", j)
    pv1 = D_[:,j]
    # pv1 = pv1/np.max(pv1)

    solver = NaiiveLrfSolver(qv, D_, E_)
    P, C = solver.solve()
    c1, c2 = C[:,j]
    aq, bq = P.T
    aq1 = aq*c1
    bq1 = bq

    Dinv = np.linalg.pinv(D_)
    W = Dinv @ P
    Pe = np.sqrt(np.dot(E_**2, W**2))

    save_aq("aq-naive.dat", qv, P, Pe)

    solver = BoundedLrfSolver(qv, D_, E_)
    P_, C_, Rg, R_, L_, hK, hL, bq_bounds_, coerced_bq_ = solver.solve(debug=False)
    aq_, bq_ = P_.T

    W = Dinv @ P_
    Peb = np.sqrt(np.dot(E_**2, W**2))

    save_aq("aq-bounded-linear-error.dat", qv, P_, Peb)

    Peb_, adjust_scale = compute_corrected_error(qv, i, P, c1, c2, P_, Pe, bq_bounds_, coerced_bq_, debug=True)

    save_aq("aq-bounded-corrected-error.dat", qv, P_, Peb_)

    def plot_diff(ax1, ax2, ax3, P, Pe=None, bounds=None):
        a = P[:,0]
        b = P[:,1]

        if Pe is None:
            W = Dinv @ P
            Pe = np.sqrt(np.dot(E_**2, W**2))

        ae = Pe[:,0]
        qc, Ic, Icerr, Dc = fit_data(qv, a, ae)
        spline = UnivariateSpline(qc, Ic, s=0, ext=3)
        delta = (a - spline(qv))/ae

        ax1.plot(qv, a, label="LRF A(q)")
        ax1.plot(qc, Ic, label="IFT fit")
        ax1.legend()
        ax2.plot(qv, b, color="pink", label="LRF B(q)")
        if bounds is not None:
            for k, bound in enumerate(bounds):
                label = "B(q) Bounds" if k == 0 else None
                ax2.plot(qv, bound, ":", color="red", label=label)
        ax2.legend()

        ax3.plot(qv, delta, "o", markersize=3)

    with plt.Dp():
        from matplotlib.gridspec import GridSpec
        from DataUtils import get_in_folder

        gs = GridSpec(4,3)

        fig = plt.figure(figsize=(18,8))
        fig.suptitle("$\Delta I/\sigma$ Comparison for %s" % get_in_folder(), fontsize=20)
        ax1 = fig.add_subplot(gs[0:2,0])
        ax2 = fig.add_subplot(gs[2,0])
        ax3 = fig.add_subplot(gs[3,0])
        ax1.set_yscale("log")
        ax1.set_title("Naive LRF (propagated error)", fontsize=16)

        ax4 = fig.add_subplot(gs[0:2,1])
        ax5 = fig.add_subplot(gs[2,1])
        ax6 = fig.add_subplot(gs[3,1])
        ax4.set_yscale("log")
        title = "Bounded LRF (propagated error)"
        ax4.set_title(title, fontsize=16)

        ax7 = fig.add_subplot(gs[0:2,2])
        ax8 = fig.add_subplot(gs[2,2])
        ax9 = fig.add_subplot(gs[3,2])
        ax7.set_yscale("log")
        title = "Bounded LRF (corrected error: adjust_scale=%.3g)" % adjust_scale
        ax7.set_title(title, fontsize=16)

        plot_diff(ax1, ax2, ax3, P, Pe=Pe)
        plot_diff(ax4, ax5, ax6, P_, Pe=Peb, bounds=bq_bounds_)
        plot_diff(ax7, ax8, ax9, P_, Pe=Peb_, bounds=bq_bounds_)

        ylim1 = ax1.get_ylim()
        ylim4 = ax4.get_ylim()
        ymin = min(ylim1[0], ylim4[0])
        ymax = max(ylim1[1], ylim4[1])
        for ax in ax1, ax4, ax7:
            ax.set_ylim(ymin, ymax)

        for ax in ax5, ax8:
            ax.set_ylim(ax2.get_ylim())

        for ax in ax6, ax9:
            ax.set_ylim(ax3.get_ylim())

        fig.tight_layout()
        plt.show()

def demo(caller):
    dialog = caller.dialog
    pdata, popts = caller.get_preview_data(with_update=False)

    print("pdata.cnv_ranges=", pdata.cnv_ranges)
    paired_range = pdata.cnv_ranges[0]
    ranges = paired_range.get_fromto_list()
    print(ranges)
    f, t = ranges[0]
    sd = dialog.sd
    demo_impl(sd, f, t)



# coding: utf-8
"""
    InspectionInAngles.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkUtils import split_geometry
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable

class InspectionInAnglesDialog(Dialog):
    def __init__(self, parent, q_vector, e_index, drawn_info):
        self.q_vector = q_vector
        self.e_index = e_index
        self.drawn_info = drawn_info
        Dialog.__init__(self, parent, "Inspection in Various Angles",
            # location="lower right"
            visible=False, auto_geometry=False, geometry_cb=self.geometry_cb)

    def show(self):
        self._show()

    def geometry_cb(self):
        self.update()
        parent = self.parent
        rootx = parent.winfo_rootx()
        rooty = parent.winfo_rooty()
        W, H, X, Y = split_geometry(parent.geometry())
        w, h, x, y = split_geometry(self.geometry())
        print([rootx, rooty])
        print([W, H, X, Y], [w, h, x, y])
        # overwrite w, h because they cannot yet be updated correctly
        w, h = 800, 800
        offsetx = W - w - 100
        offsety = H - h - 100
        self.geometry("+%d+%d" % (rootx+offsetx, rooty+offsety))

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()
        bframe = Tk.Frame(body_frame)
        bframe.pack(fill=Tk.X, expand=1)
        tbframe = Tk.Frame(bframe)
        tbframe.pack(side=Tk.LEFT, fill=Tk.X, expand=1)
        pframe = Tk.Frame(bframe)
        pframe.pack(side=Tk.RIGHT, padx=50)
        self.builf_control_panel(pframe)

        figsize = (10, 8)
        fig = plt.figure(figsize=figsize)
        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, tbframe)
        self.toolbar.update()
        self.ax = fig.gca()
        self.draw_decomposition()

    def draw_decomposition(self, e_index=None):
        ax = self.ax
        ax.cla()
        title, data, e_index_init, start, detail = self.drawn_info
        if e_index is None:
            e_index = e_index_init
        ax.set_title(title, fontsize=20)
        self.parent.solver.draw_components_with_bands(ax, data, e_index, start=start, detail=detail)
        self.mpl_canvas.draw()

    def builf_control_panel(self, pframe):
        space = Tk.Frame(pframe, width=50)
        space.pack(side=Tk.LEFT)

        label = Tk.Label(pframe, text="Q: ")
        label.pack(side=Tk.LEFT)
        self.q_value = Tk.StringVar()
        self.q_value.set('%.3g' % self.q_vector[self.e_index])
        label = Tk.Label(pframe, textvariable=self.q_value, width=6, justify=Tk.CENTER)
        label.pack(side=Tk.LEFT)

        label = Tk.Label(pframe, text="Angular index: ")
        label.pack(side=Tk.LEFT)
        self.angular_index = Tk.IntVar()
        self.angular_index.set(self.e_index)
        spinbox = Tk.Spinbox(pframe, textvariable=self.angular_index,
                                from_=0, to=len(self.q_vector)-1, increment=10,
                                justify=Tk.CENTER, width=6 )
        spinbox.pack(side=Tk.LEFT)
        self.angular_index.trace('w', self.angular_index_tracer)

    def angular_index_tracer(self, *args):
        index = self.angular_index.get()
        try:
            self.draw_decomposition(index)
            self.q_value.set('%.3g' % self.q_vector[index])
        except:
            pass



"""
    MdaDemo.py

    Copyright (c) 2020-2021, SAXS Team, KEK-PF
"""
import os
import numpy as np
import glob
from shutil import move
import MDAnalysis as mda
from MDAnalysisTests.datafiles import PSF, DCD
from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry
from molass_legacy._MOLASS.SerialSettings import get_setting
import matplotlib.pyplot as plt
# import molass_legacy.KekLib.DebugPlot as plt

def make_folders():
    md_folder = os.path.join(get_setting('analysis_folder'), 'MD')
    pdb_folder = os.path.join(md_folder, 'pdb')
    mrc_folder = os.path.join(md_folder, 'mrc')
    sas_folder = os.path.join(md_folder, 'sas')
    for folder in [pdb_folder, mrc_folder, sas_folder]:
        mkdirs_with_retry(folder)
    return pdb_folder, mrc_folder, sas_folder

def make_pdb_files():
    pdb_folder = make_folders()[0]

    num_files = len(os.listdir(pdb_folder))
    print(num_files)
    if num_files > 80:
        return

    u = mda.Universe(PSF, DCD)
    protein = u.select_atoms('protein')

    Rgyr = []
    for k, ts in enumerate(u.trajectory):
        Rgyr.append((u.trajectory.time, protein.radius_of_gyration()))
        file = os.path.join(pdb_folder, '%03d.pdb' % k)
        print('writing', file)
        protein.write(file)
    Rgyr = np.array(Rgyr)

    import molass_legacy.KekLib.DebugPlot as dplt
    fig, ax = dplt.subplots()
    ax.plot(Rgyr[:,0], Rgyr[:,1], 'r--', lw=2, label=r"$R_G$")
    ax.set_xlabel("time (ps)")
    ax.set_ylabel(r"radius of gyration $R_G$ ($\AA$)")
    # ax.figure.savefig("Rgyr.pdf")
    dplt.show()

def make_mrc_files():
    from molass_legacy.DENSS.DenssUtils import run_pdb2mrc

    pdb_folder, mrc_folder = make_folders()[0:2]
    num_files = len(os.listdir(mrc_folder))
    print(num_files)
    if num_files > 80:
        return

    for path in glob.glob(pdb_folder + r'\*.pdb'):
        mrc_file = run_pdb2mrc(path)
        print(mrc_file)
        move(mrc_file, mrc_folder)

def make_scattering_curves_from_mrc():
    import mrcfile
    from molass_legacy.Saxs.ReciprocalData import ReciprocalData

    mrc_folder, sas_folder = make_folders()[1:3]
    num_files = len(os.listdir(sas_folder))
    print(num_files)
    if num_files > 80:
        return

    q = np.linspace(0, 0.5, 500)
    for path in glob.glob(mrc_folder + r'\*.mrc'):
        print(path)
        with mrcfile.open(path) as mrc:
            data = mrc.data
        rdata = ReciprocalData(data.shape)
        F = rdata.get_reciprocal(data)
        curve = rdata.get_scattering_curve(q, F)
        error = curve*0.01
        folder, file = os.path.split(path)
        out_file = os.path.join(sas_folder, file.replace('_pdb.mrc', '.dat'))
        np.savetxt(out_file, np.array([q, curve, error]).T)

def make_scattering_curves_from_pdb():
    from SerialAtsasTools import CrysolExecutor

    pdb_folder, _, sas_folder = make_folders()
    num_files = len(os.listdir(sas_folder))
    print(num_files)
    if num_files > 80:
        return

    sas_denss_folder = sas_folder.replace('sas', 'sas-denss')

    crysol = CrysolExecutor()

    q = np.linspace(0, 0.5, 500)
    for path in glob.glob(pdb_folder + r'\*.pdb'):
        if path.find('_centered.pdb') > 0:
            continue
        print(path)
        folder, file = os.path.split(path)
        exp_file = os.path.join(sas_denss_folder, file.replace('.pdb', '.dat'))
        print('crysol.execute')
        crysol.execute(path, exp_file)
        crysol.move(path, sas_folder)

import matplotlib.animation as animation
from molass_legacy.KekLib.OurTkinter import Tk, Dialog

class SasAnimation(Dialog):
    def __init__(self, parent, sas_files):
        self.sas_files = sas_files
        Dialog.__init__(self, parent, "SasAnimation", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
        from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar

        cframe = Tk.Frame(body_frame)
        cframe.pack()
        self.fig = fig = plt.figure()
        self.mpl_canvas = FigureCanvasTkAgg(self.fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.toolbar = NavigationToolbar(self.mpl_canvas, cframe)
        self.toolbar.update()
        self.make_animation(fig)
        self.fig.tight_layout()
        self.mpl_canvas.draw()

    def make_animation(self, fig):
        from molass_legacy.KekLib.NumpyUtils import np_loadtxt
        sas_files = self.sas_files
        num_frames = len(sas_files)

        ax = fig.gca()

        ax.set_title("Scattering Curve Variation corresponding to Trajectory Variation", fontsize=16)
        ax.set_yscale('log')

        def get_data(k):
            path = sas_files[k]
            print(path)
            # data = np.loadtxt(path)
            data, _ = np_loadtxt(path)
            return data

        data = get_data(0)
        qv = data[:,0]
        max_intensity = data[0,1]

        line_init, = ax.plot(data[:,0], data[:,1], ':')
        line, = ax.plot(data[:,0], data[:,1])

        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin/2, ymax*2)

        tx = (xmin+xmax)/2
        ty = np.sqrt(ymin*ymax)
        text = ax.text(tx, ty, str(0), fontsize=50, alpha=0.1, ha='center', va='center')

        def init():
            # initialize an empty list of cirlces
            return [line_init, line, text]

        def animate(i):
            data = get_data(i)
            y = data[:,1]*max_intensity/data[0,1]
            line.set_data(qv, y)
            text.set_text(str(i))
            return [line_init, line, text]

        fig.tight_layout()

        self.anim = animation.FuncAnimation(fig, animate, init_func=init,
                                       frames=num_frames, interval=200, blit=True)
        # plt.show()

def make_animation(parent):
    sas_folder = make_folders()[2]
    sas_files = glob.glob(sas_folder + r'\*.dat')
    if len(sas_files) == 0:
        sas_files = glob.glob(sas_folder + r'\*.int')

    sas = SasAnimation(parent, sas_files)
    sas.show()

def demo(parent=None, use_denss=False):
    make_pdb_files()
    if use_denss:
        make_mrc_files()
        make_scattering_curves_from_mrc()
        make_animation(parent)
    else:
        make_scattering_curves_from_pdb()
        make_animation(parent)



"""
    Solvers.UltraNest.ProressCanvas.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import os
import numpy as np
from multiprocessing import Process, Queue

def create_server(nrows, ncols, update_queue, parent=None, caller_pid=None):
    import tkinter
    import matplotlib.pyplot as plt
    from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg

    class ProressCanvasServer:
        def __init__(self, nrows, ncols, grid_unit=0.2, parent=None):
            self.parent = parent
            fig, ax = plt.subplots(figsize=(ncols*grid_unit, nrows*grid_unit))
            fig.tight_layout()
            self.mpl_canvas = FigureCanvasTkAgg(fig, parent)
            self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
            self.mpl_canvas_widget.pack(fill=tkinter.BOTH, expand=1)
            ax.set_axis_off()
            self.fig = fig
            self.ax = ax
            self.colormap = 'Purples'
            self.parent.protocol("WM_DELETE_WINDOW", self.quit)
            print("server created")
 
        def begin_poll(self, interval=500):
            print("begin_poll")
            self.interval = interval
            self.parent.after(interval, self.poll)

        def poll(self):
            print("polling")
            while True:
                try:
                    data = update_queue.get(False)
                    if np.isscalar(data) and data == -1:
                        # stop polling
                        print("stop polling")
                        self.quit()
                        return
                    else:
                        self.update(data)
                except:
                    break

            self.parent.after(self.interval, self.poll)

        def update(self, data):
            # print("updating data=", data)
            self.ax.imshow(data, cmap=self.colormap)
            self.mpl_canvas.draw_idle()

        def quit(self):
            print("ProressCanvasServer: quit")
            self.parent.quit()

    begin_loop = False
    if parent is None:
        parent = tkinter.Tk()
        caller_text = "" if caller_pid is None else " from caller pid=%d" % caller_pid
        parent.title("Proress Canvas Server pid=%d%s" % (os.getpid(), caller_text))
        begin_loop = True
    server = ProressCanvasServer(nrows, ncols, parent=parent)
    server.begin_poll()
    if begin_loop:
        parent.mainloop()

class ProressCanvasClient:
    def __init__(self, nrows, ncols):
        self.update_queue = Queue()
        # the server must run as a separate process to avoid matplotlib "main thread is not in main loop" problem
        args = (nrows, ncols, self.update_queue)
        kwargs = dict(caller_pid=os.getpid())
        self.server_process = Process(target=create_server, args=args, kwargs=kwargs)
        self.server_process.start()

    def update(self, data):
        self.update_queue.put(data)
    
    def quit(self):
        self.update_queue.put(-1)
        self.server_process.join()

if __name__ == '__main__':
    import sys
    sys.path.append(os.path.dirname(os.path.dirname((os.path.dirname(__file__)))))
    from time import sleep

    shape = (10, 40)

    canvas = ProressCanvasClient(*shape)
    sleep(1)
    data = np.zeros(shape)
    data[5,:] = 1
    data[:,20] = 1
    canvas.update(data)

    sleep(1)
    data = np.zeros(shape)
    canvas.update(data)

    sleep(1)
    canvas.quit()



"""
    CharacteristicDemo.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import numpy as np
import time
from DataUtils import get_in_folder

def demo(in_folder, sd):
    import molass_legacy.KekLib.DebugPlot as plt
    from importlib import reload
    import Models.Characteristic
    reload(Models.Characteristic)    
    from molass_legacy.Models.Characteristic import CfSpace
    from molass_legacy.Models.CfsEvalPeaks import get_cfs_evaluated_peaks
    from molass_legacy.Models.ElutionCurveModels import egha

    print(in_folder)
    ecurve = sd.get_xray_curve()
    x = ecurve.x
    y = ecurve.y

    cfs = CfSpace(use_np_fft=True)
    t0 = time.perf_counter()
    w = cfs.get_w()
    cft = cfs.compute_cf(x, y)
    t1 = time.perf_counter()

    print("it took:", t1 - t0)
    print("len(cft)=", len(cft))

    num_peaks = len(ecurve.peak_info)
    egha_params, emga_params = get_cfs_evaluated_peaks(x, y, num_peaks, debug=False)

    cy_list = []
    for p in egha_params:
        cy = egha(x, *p)
        cy_list.append(cy)
    ty = np.sum(cy_list, axis=0)

    ty_cft = cfs.compute_cf(x, ty)

    in_folder_ = get_in_folder(in_folder)
    with plt.Dp():
        fig, axes = plt.subplots(nrows=2, ncols=4, figsize=(20,10))
        fig.suptitle("Decomposition Demo using CF Space with %s" % in_folder_, fontsize=20)
        ax1, ax2, ax3, ax4 = axes[0,:]

        ax1.set_title("Data", fontsize=16)
        ax2.set_title("cft.real vs w", fontsize=16)
        ax3.set_title("cft.imag vs w", fontsize=16)
        ax4.set_title("cft.imag vs cft.real", fontsize=16)

        ax1.plot(x, y)

        ax2.plot(w, cft.real)
        ax3.plot(w, cft.imag)
        ax4.plot(cft.real, cft.imag, "o", markersize=3)

        ax5, ax6, ax7, ax8 = axes[1,:]

        ax5.plot(x, y)
        for cy in cy_list:
            ax5.plot(x, cy, ":")
        ax5.plot(x, ty, ":", lw=3, color="red")

        ax6.plot(w, ty_cft.real, color="red")
        ax7.plot(w, ty_cft.imag, color="red")
        ax8.plot(ty_cft.real, ty_cft.imag, "o", markersize=3, color="red")

        fig.tight_layout()
        plt.show()



"""
    Estimators.RtEmgEstimator.py

    Copyright (c) 2022-2023, SAXS Team, KEK-PF
"""
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.Peaks.ElutionModels import emg
from .EghEstimator import EghEstimator

class RtEmgEstimator(EghEstimator):
    def __init__(self, editor):
        EghEstimator.__init__(self, editor)

    def estimate_params(self, debug=True):
        if debug:
            from importlib import reload
            import SecTheory.HermansEmg
            reload(SecTheory.HermansEmg)
        from SecTheory.HermansEmg import estimate_initial_params, convert_to_xr_params_hermans

        init_xr_params, init_xr_baseparams, temp_rgs, init_mapping, init_uv_heights, init_uv_baseparams, init_mappable_range, seccol_params = self.estimate_egh_params()

        D = self.editor.dsets[0][1]
        temp_xr_params, seccol_params, R = estimate_initial_params(self.ecurves[1], init_xr_params, temp_rgs, seccol_params, D)

        baseline_type = get_setting("unified_baseline_type")
        if baseline_type >= 2:
            # recompute uv_baseparams
            xr_params = convert_to_xr_params_hermans(temp_xr_params, temp_rgs, seccol_params, R)
            uv_curve, xr_curve = self.ecurves
            x = xr_curve.x
            a, b = init_mapping
            uv_x = a*x+b
            uv_y = uv_curve.spline(uv_x)
            uv_ty = np.zeros(len(uv_x))
            k = 0
            for h, m, s, t in xr_params:
                uv_cy = emg(uv_x, init_uv_heights[k], a*m+b, a*s, a*t)
                uv_ty += uv_cy
                k += 1

            uv_base_params = self.editor.get_uv_base_params(xyt=(uv_x, uv_y, uv_ty))
        else:
            uv_base_params = init_uv_baseparams

        return np.concatenate([temp_xr_params.flatten(), init_xr_baseparams, temp_rgs, init_mapping, init_uv_heights, uv_base_params, init_mappable_range, seccol_params, [R]])



"""
    Estimators.StcEstimator.py

    Copyright (c) 2022-2025, SAXS Team, KEK-PF
"""
import numpy as np
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.SecTheory.RetensionTime import estimate_init_rgs
from molass_legacy.Peaks.PeProgressConstants import MAXNUM_STEPS, STOCH_INIT_STEPS
from .BaseEstimator import BaseEstimator

class StcEstimator(BaseEstimator):
    def __init__(self, editor):
        BaseEstimator.__init__(self, editor)

    def estimate_params(self, debug=False):
        return self.compute_stochastic_init_params(self.nc, self.t0_upper_bound, debug=debug)

    def compute_stochastic_init_params(self, nc_b, t0_upper_bound, debug=False):
        if debug:
            from importlib import reload
            import molass_legacy.SecTheory.MonoPore
            reload(molass_legacy.SecTheory.MonoPore)
            import molass_legacy.Models.Stochastic.MonoporeUvScaler
            reload(molass_legacy.Models.Stochastic.MonoporeUvScaler)
        from molass_legacy.SecTheory.MonoPore import estimate_monopore_params, get_modified_params, split_params, estimate_uv_scale_params
        from molass_legacy.Models.Stochastic.MonoporeUvScaler import adjust_to_uv_scales

        editor = self.editor

        progress = MAXNUM_STEPS - STOCH_INIT_STEPS
        editor.update_status_bar("Estimating stochastic initial parameters.")
     
        nc = nc_b - 1   # num components without baseline
        (xr_curve, D), rg_curve, (uv_curve, U) = editor.dsets

        lrf_src = editor.get_lrf_source(devel=True)
        if lrf_src is None:
            ret = estimate_monopore_params(xr_curve, rg_curve, nc, optimizer=editor.fullopt,
                                           t0_upper_bound=t0_upper_bound, logger=self.logger, debug=debug)
            init_params_for_uv = ret.x
            monopore_params, rgs, xr_w = split_params(ret.x, nc)
        else:
            ret = lrf_src.guess_monopore_params(debug=debug)
            if ret is None:
                return
            xr_init_params, corrected_rgs = ret
            N, T, x0, me, mp, poresize = xr_init_params[0:6]
            xr_w = xr_init_params[6:]
            rgs = corrected_rgs
            monopore_params = np.array([x0, poresize, N, me, T, mp])    # t0, rp, N, me, T, mp

        progress += 1
        editor.pbar["value"] = progress
        editor.update()

        _, _, xr_x, xr_y, baselines = editor.get_curve_xy(return_baselines=True)
        a, b = editor.peak_params_set[-2:]
        uv_x = xr_x*a + b
        uv_y = uv_curve.spline(uv_x)

        uv_baseline = editor.get_uv_baseline(xy=(uv_x, uv_y))

        uv_y_ = uv_y - uv_baseline
        xr_y_ = xr_y - baselines[1]

        ret = adjust_to_uv_scales(xr_x, xr_y_, uv_x, uv_y_, xr_init_params, corrected_rgs, debug=debug)
        if ret is None:
            return
        else:
            uv_w, uv_ty = ret

        # uv_w, uv_ty = estimate_uv_scale_params(xr_curve, rg_curve, uv_curve, nc, init_params_for_uv, xr_w, uv_x, uv_y_, xr_x, xr_y_, optimizer=editor.fullopt, debug=debug)

        uv_base_params = editor.get_uv_base_params(xyt=(uv_x, uv_y, uv_ty))

        if debug:
            self.logger.info("uv_base_params=%s, average(uv_ty)=%g", str(uv_base_params), np.average(uv_ty))
            uv_bl_computed = editor.base_curve_info[0](uv_x, uv_base_params, uv_ty)
            with plt.Dp():
                fig,(ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("compute_stochastic_init_params")
                ax1.plot(uv_x, uv_y)
                ax1.plot(uv_x, uv_ty + uv_baseline, ":")
                ax1.plot(uv_x, uv_baseline, color="red")
                ax1.plot(uv_x, uv_bl_computed, color="yellow")
                ax2.plot(xr_x, xr_y)
                ax2.plot(xr_x, baselines[1], color="red")
                fig.tight_layout()
                plt.show()

        param_list = [xr_w, editor.baseline_params[1], rgs, (a, b), uv_w, uv_base_params, xr_x[[0,-1]], monopore_params]
        init_params = np.concatenate(param_list)

        progress += 1
        editor.pbar["value"] = progress
        editor.update_status_bar("Stochastic initial parameters are ready.")

        return init_params

def onthefly_test(editor):
    estimator = StcEstimator(editor)
    print("estimating...")
    init_params = estimator.estimate_params(debug=True)
    print("done.")
    if init_params is not None:
        editor.draw_scores(init_params)
        print("redraw done.")


"""
    SecTheory.ColumnIllust.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
import os
import sys
import numpy as np
from matplotlib.gridspec import GridSpec
from scipy.interpolate import UnivariateSpline
from matplotlib.patches import Rectangle, Circle

R = 1/12
dr = R*0.3
r = R*0.7
width = 2*r
height = 2*r
poresize_max = 200

def plot_column_illust(ax, poresizes, colors):
    ax.set_axis_off()

    num_iters = 36//len(poresizes)
    poreradii = np.array(poresizes * num_iters)
    porecolors = np.array(colors * num_iters)
    print(len(poreradii))
    indeces = np.arange(36)
    np.random.shuffle(indeces)
    print(poreradii[indeces])

    patches = []
    k = 0
    for i in range(6):
        for j in range(6):
            rectangle = Rectangle((dr+i/6, dr+j/6), width, height, facecolor="black")
            ax.add_patch(rectangle)
            k_ = indeces[k]
            pore = Circle((R+i/6, R+j/6), r*poreradii[k_]/poresize_max, color=porecolors[k_])
            ax.add_patch(pore)
            k += 1

def plot_column_legend(ax, poresizes, colors):
    ax.set_axis_off()

    dy = 0.2
    x = 0.2

    for k, (size, color) in enumerate(zip(poresizes, colors)):
        y = 0.7 - k*R*2
        rectangle = Rectangle((dr+x, dr+y), width, height, facecolor="black")
        ax.add_patch(rectangle)
        pore = Circle((R+x, R+y), r*size/poresize_max, color=color)
        ax.add_patch(pore)
        ax.text(R+x+0.2, R+y, "Poresize: %3.0f" % size, va="center", fontsize=16)

def demo():
    import molass_legacy.KekLib.DebugPlot as plt

    fig, (ax1, ax2, ax3, ax4) = plt.subplots(ncols=4, figsize=(18,4.5))

    plot_column_illust(ax1, [76], ["white"])
    plot_column_illust(ax2, [90, 40], ["orange", "yellow"])
    plot_column_illust(ax3, [100, 76, 40], ["pink", "white", "yellow"])
    plot_column_legend(ax4, [100, 90, 76, 40], ["pink", "orange", "white", "yellow"])

    fig.tight_layout()
    plt.show()

if __name__ == '__main__':
    this_dir = os.path.dirname(os.path.abspath(__file__))
    lib_dir = os.path.dirname(this_dir)
    sys.path.append(lib_dir)

    import seaborn
    seaborn.set()
    import molass_legacy.KekLib

    demo()



"""
    UV.UvBaseCurve.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import minimize, basinhopping
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.UV.PlainCurve import make_secondary_e_curve_at
from molass_legacy.Trimming.Sigmoid import guess_bent_sigmoid, ex_sigmoid
from molass_legacy.Peaks.EghSupples import egh, d_egh
from molass_legacy.KekLib.SciPyCookbook import smooth

SMOOTHING = True

class UvBaseCurve:
    def __init__(self, sd, curve1, n_sigmchain, n_peaks):
        self.curve1 = curve1
        pn = curve1.primary_peak_no
        emg_peaks = curve1.get_emg_peaks()
        epk = emg_peaks[pn]
        self.peak_params = epk.get_params()
        self.n_sigmchain = n_sigmchain
        self.n_peaks = n_peaks

    def guess_params(self, curve2, pp, slice_, debug=False):
        x = curve2.x
        y = curve2.y

        x_ = x[slice_]
        if SMOOTHING:
            y_ = smooth(y[slice_])
        else:
            y_ = y[slice_]

        def obj_func(p):
            fy = self.__call__(x_, p)
            return np.sum((fy - y_)**2)

        x0 = pp[0]
        start = max(0, x0 - 50)
        stop = min(len(x), x0 + 50)
        guess_slice = slice(start, stop)
        sigm_params = guess_bent_sigmoid(x[guess_slice], y[guess_slice], x0=x0)

        opt_ret = None
        for scale in [-1, 1]:
            init_params = (*sigm_params, scale)
            ret = minimize(obj_func, init_params)
            # ret = basinhopping(obj_func, init_params)
            print("ret.x=", ret.x)
            if opt_ret is None or ret.fun < opt_ret.fun:
                opt_ret = ret

        if debug:
            with plt.Dp():
                from molass_legacy.Elution.CurveUtils import simple_plot
                fig, (ax1, ax2) = plt.subplots(ncols=2, figsize=(12,5))
                fig.suptitle("debug plot at guess_params")
                simple_plot(ax1, self.curve1)
                ax1.plot(x, egh(x, *self.peak_params))
                ax2.plot(x, y)
                if SMOOTHING:
                    ax2.plot(x_, y_)
                fy = self.__call__(x, opt_ret.x)
                ax2.plot(x, fy)
                ax3.plot()
                fig.tight_layout()
                plt.show()

        return opt_ret.x

    def __call__(self, x, params):
        """
            requires 6 + 1 params when n_sigmchain=1, n_peaks=1
        """
        y = np.zeros(len(x))
        if self.n_sigmchain == 1:
            start = 6
            y += ex_sigmoid(x, *params[0:start])
        else:
            # SigmoidChain
            assert False
        h, m, s, t = self.peak_params
        for k in range(self.n_peaks):
            k_ = start + k
            y += d_egh(x, h*params[k], m, s, t)
        return y



"""
    Alsaker.Bridge.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import os
import numpy as np
from testfixtures import replace_in_module
import rpy2
import rpy2.robjects as robjects
from rpy2.robjects.packages import STAP
from rpy2.robjects import default_converter, numpy2ri

R_read_table = robjects.r['read.table']
R_matrix = robjects.r['matrix']
R_rep = robjects.r['rep']

"""
    These are patch functions to suppress decode errors which are ignorable.
"""
import rpy2.rinterface_lib.conversion as conversion
from rpy2.rinterface_lib.conversion import ffi

def modified_cchar_to_str(c, encoding: str) -> str:
    # TODO: use isString and installTrChar
    s = ffi.string(c).decode(encoding, errors='ignore')
    return s

def modified_cchar_to_str_with_maxlen(c, maxlen: int, encoding: str) -> str:
    # TODO: use isString and installTrChar
    s = ffi.string(c, maxlen).decode(encoding, errors='ignore')
    return s

# end of patch functions

def do_it_in_the_context(closure):
    """
    workaround for multiple patching of rpy2
    i.e., generator tuple or list are not supported in with statement
    """
    with (  replace_in_module(conversion._cchar_to_str, modified_cchar_to_str),
            replace_in_module(conversion._cchar_to_str_with_maxlen, modified_cchar_to_str_with_maxlen)
    ):
        return closure()

class RgcodeBrige:
    def __init__(self):
        this_dir = os.path.dirname(__file__)
        self.rg_code_dir = os.path.join(this_dir, "Rg code")
        rg_code_file = os.path.join(self.rg_code_dir, "file1-modified-for-molass.R")
        print("R_HONE=", os.environ['R_HOME'])
        print(rpy2.__version__)
        with open(rg_code_file) as f:
            rg_code_string = f.read()

        self.rg_code = STAP(rg_code_string, "estimate_Rg")
 
    def get_sample_data_path(self, name):
        return os.path.join(self.rg_code_dir, name)

    def estimate_Rg(self, data, num_reps, starting_value=None, return_y3=False):
        if starting_value is None:
            starting_value = R_rep(1, num_reps)
        else:
            starting_value = R_rep(starting_value, num_reps)

        print("type(starting_value)=", type(starting_value))    # <class 'rpy2.robjects.vectors.IntVector'>
        print("starting_value=", starting_value)

        if type(data) is np.ndarray:
            numpy2ri.activate()

            nr, nc = data.shape
            M = R_matrix(data, nrow=nr, ncol=nc)
            # this call must take place before numpy2ri.deactivate()
            output = self.rg_code.estimate_Rg(M, num_reps, starting_value=starting_value, return_y3=return_y3)

            numpy2ri.deactivate()
        else:
            M = data
            output= self.rg_code.estimate_Rg(M, num_reps, starting_value=starting_value)
        return output

def bridge_test_impl(caller):
    print("Bridge trial")

    def closure():
        rb = RgcodeBrige()
        filepath = rb.get_sample_data_path("oval_01C_S008_0_01.dat")

        # basic call test
        data =  R_read_table(filepath, header=False)
        print(type(data))
        output = rb.estimate_Rg(data, 1)
        print(type(output), output)

        # numpy interface test
        data = np.loadtxt(filepath)
        print("data.shape=", data.shape)
        output = rb.estimate_Rg(data, 1)
        print(type(output), output)
    
    do_it_in_the_context(closure)


"""
    PreliminaryRg.py

    Copyright (c) 2019-2025, SAXS Team, KEK-PF
"""

import logging
import numpy as np
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.GuinierAnalyzer.SimpleGuinier import SimpleGuinier
from molass_legacy.Trimming import FlangeLimit, GuinierLimit, CdLimit
import molass_legacy.KekLib.DebugPlot as plt

class PreliminaryRg:
    def __init__(self, D, E, e_curve, qv, flange_limit, **kwargs):
        self.logger = logging.getLogger(__name__)
        self.qv = qv
        self.default_eno = e_curve.get_primarypeak_i()
        self.D = D
        self.E = E
        self.flange_limit = flange_limit
        self.current_eno = None
        self.compute_rg(**kwargs)
        self.logger.info("default_eno=%d, current_eno=%d", self.default_eno, self.current_eno)

    def compute_rg(self, selected=None, ip_effect_info=None, debug=False):
        if selected is None:
            selected = self.default_eno

        if self.current_eno is None or selected != self.current_eno:
            qslice = slice(0, self.flange_limit)
            scattering_y, selected = self.get_scattering_y(selected, qslice, return_selection=True)
            data = np.vstack( [self.qv[qslice], scattering_y, self.E[qslice, selected]] ).T
            self.sg = sg = SimpleGuinier( data )
            """
            self.sg is going to be used in AngularRange.GuinierLimit,
            which may not be safe for erroneous cases below
            """
            if sg.Rg is None or sg.Rg < 10.0:
                self.Rg = 10.0
                self.guinier_start = 0
                self.logger.warning("Rg=%s < 10.0 has been replaced by 10.0 to avoid exception.", str(sg.Rg))
            else:
                self.Rg = sg.Rg
                self.guinier_start = sg.guinier_start
            self.logger.info("preliminary Rg=%.3g at eno=%d", self.Rg, selected)
            self.current_eno = selected

            if debug:
                print("self.guinier_start=", self.guinier_start )
                qv = self.qv
                glim = sg.guinier_stop
                qv2 = qv[0:glim]**2
                with plt.Dp():
                    fig, ax = plt.subplots()
                    ax.set_title("PreliminaryRg.compute_rg debug")
                    ax.plot(qv2, np.log(scattering_y[0:glim]))
                    ax.plot(sg.guinier_x, sg.guinier_y)
                    fig.tight_layout()
                    plt.show()

        return self.Rg

    def get_scattering_y(self, selected=None, qslice=None, return_selection=False):
        if selected is None:
            selected = self.default_eno
        if qslice is None:
            qslice = slice(0, self.flange_limit)
        start = max(0, selected - 2)
        stop  = min(self.D.shape[1], selected + 2)
        y = np.average(self.D[qslice, start:stop], axis=1)
        if return_selection:
            return y, selected
        else:
            return y

    def get_guinier_start_index(self, **kwargs):
        self.compute_rg(**kwargs)
        return self.guinier_start

def get_flange_limit(D, E, e_curve, qv):
    fl = FlangeLimit(D, E, e_curve, qv)
    return fl.get_limit()

def get_small_anlge_limit(pre_rg, D, E, e_curve, qv, qlimit, logger):

    cut_before_guinier = get_setting('cut_before_guinier')
    if cut_before_guinier == 0:
        angle_start = 0
    else:
        guinier_start = pre_rg.get_guinier_start_index()
        if cut_before_guinier == 1:
            angle_start = guinier_start
            if angle_start > 0:
                logger.info('angular range start Q[%d]=%g has been set from the Guinier interval.' % (angle_start, qv[angle_start]) )
        else:
            gl = GuinierLimit(D, e_curve, pre_rg, qlimit)
            rg_consistency = get_setting('acceptable_rg_consist')
            try:
                angle_start = gl.get_limit(rg_consistency, debug=False)
            except:
                from molass_legacy.KekLib.ExceptionTracebacker import log_exception
                log_exception(logger, "GuinierLimit.get_limit failed.", n=10)
                angle_start = 0

            if angle_start > 0:
                logger.info('angular range start Q[%d]=%g has been set from the extended Guinier limit.' % (angle_start, qv[angle_start]) )

            use_bqlimit = get_setting('use_bqlimit')
            if use_bqlimit:
                try:
                    bl = CdLimit(D, E, e_curve, qv)
                    b_limit, surely = bl.get_limit()
                    if b_limit > angle_start:
                        logger.info("Guinier limit %d has been superseded by B(q) limit %d.", angle_start, b_limit)
                        angle_start = b_limit
                    elif b_limit < angle_start:
                        force_bqlimit = get_setting('force_bqlimit')
                        if force_bqlimit and surely:
                            logger.info("Guinier limit %d has been surely replaced by B(q) limit %d.", angle_start, b_limit)
                            angle_start = b_limit
                except:
                    from molass_legacy.KekLib.ExceptionTracebacker import log_exception
                    log_exception(logger, "getting B(q) limit failed")

        return angle_start

def get_default_angle_range_impl(D, E, e_curve, qv, qlimit, logger):
    flange_limit =get_flange_limit(D, E, e_curve, qv)

    pre_rg = PreliminaryRg(D, E, e_curve, qv, flange_limit)

    angle_start = get_small_anlge_limit(pre_rg, D, E, e_curve, qv, qlimit, logger)

    return angle_start, flange_limit, pre_rg



"""
    ExtrapolSolverDialog.py

    Copyright (c) 2018-2024, SAXS Team, KEK-PF
"""
import os
import copy
import re
from bisect import bisect_right
import logging
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
import matplotlib.patches as mpl_patches
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkUtils import adjusted_geometry
from molass_legacy.KekLib.TkSupplements import BlinkingFrame
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar, get_color
from molass_legacy.KekLib.DebugPlot import set_plot_env
from molass_legacy.KekLib.ExceptionTracebacker import log_exception
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting
from .ExtrapolationControlPanel import ControlPanel, ENABLE_KNOWN_INPUT
from .ExtrapolationAnimation import TOO_SMALL_TO_PLOT
from .ExtrapolationDenssMenu import ExtrapolationDenssMenu
from molass_legacy.Test.TesterLogger import write_to_tester_log
from molass_legacy.SerialAnalyzer.DevSettings import set_dev_setting
from molass_legacy.SerialAnalyzer.DataUtils import compact_path_name
from .PreviewButtonFrame import ELUTION_DATATYPE_NAMES
from molass_legacy.Selective.AdvancedFrame import get_change_name
from molass_legacy._MOLASS.Version import is_developing_version

SMALL_ANGLE_LIMIT_Q = 0.1
B_SHIFT_POINT_Q     = 0.02
AUTO_CDI_BUTTON = False

def format_func(value, tick_number):
    return r'$10^{%d}$' % (value)

class ExtrapolSolverDialog( Dialog ):
    def __init__( self, parent, pool, editor_frame=None, last_change_id=None, scrollable=True ):
        self.mplt_ge_2_2 = matplotlib.__version__ >= '2.2'
        self.logger = logging.getLogger(__name__)
        self.grab = 'local'     # used in grab_set
        self.parent = parent
        set_plot_env(sub_parent=self.parent)    # for debug plot in case when set_plot_env has not yet called, e.g., running the while app.
        pdata = pool.pdata
        self.pdata = pdata
        self.popts = pool.popts
        self.mapper = pdata.mapper
        self.editor_frame = editor_frame
        self.last_change_id = last_change_id

        self.sd = pdata.sd
        self.j0 = pdata.sd.xr_j0
        self.conc = pdata.conc
        self.conc_curves = pdata.conc_curves
        self.elution_y =  pdata.make_conc_vector()
        self.mc_vector  = pdata.mc_vector
        self.doing_sec = pdata.is_for_sec
        self.arg_paired_ranges = pdata.paired_ranges    # keep this to call resursively in solve_unknowns
        self.cnv_ranges = pdata.cnv_ranges
        self.num_ranges = pdata.num_ranges
        self.known_info_list = None
        self.rank_increment = get_setting('rank_increment')

        solver = pool.solver
        self.solver = solver
        self.data   = solver.data
        self.error  = solver.error
        self.x_curve = solver.ecurve
        self.aq_smoothness = solver.aq_smoothness
        self.aq_positivity = solver.aq_positivity

        if pdata.is_for_sec:
            self.q = pdata.sd.intensity_array[0,:,0]
        else:
            self.q = pdata.xdata.vector

        self.selector = pool.selector
        self.solver_results = pool.get_better_results()
        self.use_elution_models = pool.popts.use_elution_models
        self.lrf_bound_correction = get_setting("lrf_bound_correction")
        self.scrollable = scrollable
        self.DEFAULT_WEIGHTS = get_setting('penalty_weighting')
        self.applied    = False
        self.geometry_init = True
        Dialog.__init__( self, self.parent, "Extrapolation Preview", visible=False)

    def cancel(self):
        # overiding cancel to cleanup self.fig
        # because the call to the destructor __del__ seems to be delayed
        plt.close(self.fig)
        # print("ExtrapolSolverDialog: closed fig")
        Dialog.cancel(self)

    def show( self ):
        self._show()

    def body( self, body_frame ):

        title_frame = Tk.Frame( body_frame )
        title_frame.pack()
        in_folder_name = compact_path_name(get_setting('in_folder'))
        concentration_datatype = get_setting('concentration_datatype')
        data_type_name = ELUTION_DATATYPE_NAMES[concentration_datatype]
        preview_model = get_setting('preview_model')
        if preview_model is not None:
            data_type_name = data_type_name.replace("elution", preview_model)

        last_change = "" if self.last_change_id is None else " after " + get_change_name(self.last_change_id)
        title = "LRF Preview of %s using %s%s" % (in_folder_name, data_type_name, last_change)
        label = Tk.Label(title_frame, text=title, font=("", 20) )
        label.pack(pady=5)

        bottom_frame = Tk.Frame(body_frame)
        bottom_frame.pack( side=Tk.BOTTOM, fill=Tk.X )
        self.bottom_frame = bottom_frame

        if self.scrollable:
            from molass_legacy.KekLib.ScrolledFrame import ScrolledFrame
            self.scrolled_frame = ScrolledFrame(body_frame)
            # self.scrolled_frame.pack( fill=Tk.BOTH, expand=1 )
            self.scrolled_frame.pack(anchor=Tk.N)
            fframe = self.scrolled_frame.interior
        else:
            self.scrolled_frame = None
            fframe = Tk.Frame( body_frame )
            fframe.pack(anchor=Tk.N)

        bottom_frame1 = Tk.Frame( bottom_frame )
        bottom_frame1.pack(fill=Tk.X, pady=5)

        self.guide_frame = Tk.Frame(bottom_frame)
        self.guide_frame.pack(fill=Tk.X)

        from molass_legacy.QuickAnalysis.AnalysisGuide import get_data_correction_state_text
        text = "Showing data processed with " + get_data_correction_state_text()
        guide_label = Tk.Label(self.guide_frame, text=text, bg='white')
        guide_label.pack(fill=Tk.X, pady=10)
        self.guide_message = guide_label

        tframe = Tk.Frame( bottom_frame1 )
        tframe.pack( side=Tk.LEFT, fill=Tk.X, expand=1 )
        space = Tk.Frame(bottom_frame1, width=50)
        space.pack(side=Tk.LEFT)
        bframe = Tk.Frame( bottom_frame1 )
        bframe.pack( side=Tk.RIGHT )

        cframe = Tk.Frame( fframe )
        cframe.pack( side=Tk.LEFT )

        self.pframe = pframe = Tk.Frame( fframe )
        pframe.pack( side=Tk.LEFT, fill=Tk.BOTH, padx=10 )

        x_curve = self.x_curve

        sd  = self.sd
        data = self.data

        # num_peaks = len( x_curve.peak_info )
        num_panels = self.num_ranges
        h = 4
        self.fig = fig = plt.figure( figsize=(20, num_panels*h) )
        gs = GridSpec(num_panels, 11)

        self.mpl_canvas = FigureCanvasTkAgg( fig, cframe )
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        self.canvas_draw()

        self.toolbar = NavigationToolbar( self.mpl_canvas, tframe )
        self.toolbar.update()

        self.fx = np.arange( data.shape[1] )
        self.elution_x = self.j0 + self.fx

        self.sa_limit = bisect_right( self.q, SMALL_ANGLE_LIMIT_Q )
        self.qs = self.q[0:self.sa_limit]
        self.b_shift_point = bisect_right( self.q, B_SHIFT_POINT_Q )

        # previous_penalty_matrix = get_setting('zx_penalty_matrix')
        # print('previous_penalty_matrix=', previous_penalty_matrix)
        previous_penalty_matrix = None
        # TODO: remove this control by previous_penalty_matrix

        self.zx_penalty_matrix = []
        self.axis_array = []
        self.super_panel_list = []
        self.panel_list = []

        ignore_all_bqs = get_setting('ignore_all_bqs')
        # ignore_bq_list = get_setting('ignore_bq_list')
        ignore_bq_list = None
        selector = self.selector

        row = 0
        for pno, range_ in enumerate(self.cnv_ranges):
            fromto_list = range_.get_fromto_list()
            for ad, _ in enumerate(fromto_list):
                pframe.grid_rowconfigure( row, weight=1 )
                peakset_info = selector.select_peakset(row)
                if previous_penalty_matrix is None:
                    weights = copy.deepcopy(self.DEFAULT_WEIGHTS)
                else:
                    weights = previous_penalty_matrix[row]
                self.zx_penalty_matrix.append(weights)
                if ignore_bq_list is None:
                    ignore_bq = ignore_all_bqs
                else:
                    ignore_bq = ignore_bq_list[row]
                panel = ControlPanel( pframe, self, pno, ad, range_, row, peakset_info, weights, ignore_bq, editor=self.editor_frame)
                panel.grid( row=row, column=0 )
                self.super_panel_list.append(panel)
                self.panel_list.append(panel.unknown_panel)

                ax1 = fig.add_subplot(gs[row, 0:2])
                ax2A1 = fig.add_subplot(gs[row, 2:5])
                ax2B1 = ax2A1.twinx()
                ax3 = fig.add_subplot(gs[row, 5:8])
                ax4 = fig.add_subplot(gs[row, 8:11])
                self.axis_array.append([ax1, ax2A1, ax2B1, ax3, ax4])
                row += 1

        self.to_solve_ranges = []
        self.Y_list = []
        self.topx_list = []
        self.sg_list = []
        self.data_list = []
        row = 0
        self.rg_quality_list = []
        quality_warning = False
        for pno, paired_range in enumerate(self.cnv_ranges):
            # print( 'paired_range=', paired_range )
            top_ = paired_range.top_x
            fromto_list = paired_range.get_fromto_list()
            for range_ in fromto_list:
                start   = range_[0]
                stop    = range_[1] + 1
                self.to_solve_ranges.append( [ paired_range, start, stop ] )
                top_c = self.mc_vector[top_]
                # top_c = elution_y[top_]
                Y   = data[:,top_]
                Y_  = Y / top_c
                Y_[Y_ < TOO_SMALL_TO_PLOT] = 0  # to avoid deforming the figure with these values
                self.Y_list.append( Y_ )
                self.topx_list.append(top_)

                peakset_info = selector.select_peakset(row)
                # print([pno], 'peakset_info=', peakset_info)

                if ignore_bq_list is None:
                    ignore_bq = ignore_all_bqs
                else:
                    ignore_bq = ignore_bq_list[row]
                self.data_list.append(None)
                lrf_info, C = self.solve_plot( row, peakset_info, init_mode=True, ignore_bq=ignore_bq )
                self.rg_quality_list.append((lrf_info.Rg, lrf_info.basic_quality))
                if lrf_info.basic_quality < 0.5:
                    quality_warning = True
                self.panel_list[row].set_C_matrix(C)
                self.sg_list.append(lrf_info.sg)
                self.plot_guinier_kratky(row)
                row += 1

        if quality_warning:
            fully_automatic = get_setting('fully_automatic')
            suppress_low_quality_warning = get_setting('suppress_low_quality_warning')
            if fully_automatic or suppress_low_quality_warning:
                pass
            else:
                # after(1000, ..) seems too early for slower machines
                self.after(2000, self.show_quality_warning)

        fig.tight_layout()
        fig.subplots_adjust(left=0.02, wspace=1.5)
        # plt.show()
        self.log_rg_info()

        self.canvas_draw()

        if AUTO_CDI_BUTTON:
            self.auto_cdi_btn = Tk.Button(bframe, text="auto-CDI for all", command=self.auto_cdi_for_all )
            self.auto_cdi_btn.pack(side=Tk.RIGHT, padx=10)

        self.denss_menu = ExtrapolationDenssMenu(bframe, self, self.data_list)
        self.denss_menu.pack(side=Tk.RIGHT, padx=30)

        if ENABLE_KNOWN_INPUT:
            self.su_btn_blink = BlinkingFrame(bframe)
            self.su_btn_blink.pack(side=Tk.RIGHT, padx=10)
            btn = Tk.Button(self.su_btn_blink, text="Solve Unknowns", command=self.solve_unknowns, state=Tk.NORMAL )
            btn.pack()
            self.su_btn_blink.objects = [btn]

        self.save_btn = Tk.Button(bframe, text="Save results", command=self.save_results )
        self.save_btn.pack(side=Tk.RIGHT, padx=10)

        self.add_dnd_bind()

        self.set_geometry()

    def show_quality_warning(self):
        if get_setting('test_pattern') is not None:
            # suppress warning while testing
            return

        import molass_legacy.KekLib.CustomMessageBox as MessageBox

        self.update()
        message =  (
                "     Rg    basic quality\n"
                "   ---------------------\n"
                )
        for rg, quality in self.rg_quality_list:
            rg_ = " None" if rg is None else "%5.1f" % rg
            low_qaulity_text = "   ← low quality" if quality < 0.5 else ""
            message += "    %s     %4.2f%s\n" % (rg_, quality, low_qaulity_text)

        if self.popts.conc_depend > 1:
            message += ("\nand that\n"
                        '"low quality" may have been caused by rank mismatch,\n'
                        "in that case, consider retrying with Conc. Dependency=1."
                        )

        MessageBox.showwarning("Guinier Analysis Quality Warning",
            "Be aware that this preview includes some low quality results\n"
            "as shown below\n\n"
            "%s" % message,
            parent=self)

    def set_geometry(self):
        from molass_legacy.KekLib.MultiMonitor import get_selected_monitor
        self.update()
        canvas_width = int(self.mpl_canvas_widget.cget( 'width' ))
        canvas_height = int(self.mpl_canvas_widget.cget( 'height' ))
        panel_width = 180
        print("canvas_height=", canvas_height)
        monitor = get_selected_monitor()
        monitor_height = monitor.height
        height = min(int(monitor_height*0.8), canvas_height+200)
        wxh = '%dx%d' % (canvas_width+panel_width, height)
        geometry = self.geometry()
        new_geometry = re.sub( r'(\d+x\d+)(.+)', lambda m: wxh + m.group(2), geometry)
        self.geometry(new_geometry)

    def canvas_draw( self ):
        if self.mplt_ge_2_2:
            self.mpl_canvas.draw()
        else:
            self.mpl_canvas.show()

    def buttonbox( self ):
        box = Tk.Frame(self.bottom_frame)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=50, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=50, pady=5)

        if is_developing_version():
            w = Tk.Button(box, text="On the Fly", width=10, command=self.on_the_fly)
            w.pack(side=Tk.LEFT, padx=50, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def solve_plot( self, row, peakset_info, penalty_weights=None, init_mode=False, animation=False, ignore_bq=0 ):
        paired_range, start, stop = self.to_solve_ranges[row]

        # A, B = self.solver.extrapolate( start, stop, with_intercept=False )
        if init_mode:
            A, B, Z, E, lrf_info, C = self.solver_results[row]
        else:
            A, B, Z, E, lrf_info, C = self.solver.extrapolate_wiser( start, stop, peakset_info,
                                ignore_bq=ignore_bq,
                                penalty_weights=penalty_weights, animation=animation )
            self.solver_results[row] = [A, B, Z, E, lrf_info, C]

        need_bq = lrf_info.need_bq()

        file_name = self.panel_list[row].get_file_name()
        self.data_list[row] = (self.q, A, E[0], file_name)
        Bs = B[self.b_shift_point]

        i = (start + stop)//2

        ax1, ax2A1, ax2B1 = self.axis_array[row][0:3]
        for ax in [ax1, ax2A1, ax2B1]:
            ax.cla()

        ax1.set_title("Range", fontsize=16)
        ax1.axes.get_xaxis().set_ticks([])
        ax1.axes.get_yaxis().set_ticks([])
        ax1.plot(self.elution_x, self.elution_y, color='blue')
        topx = self.topx_list[row]
        ax1.plot(self.j0+topx, self.elution_y[topx], 'o', color='yellow')

        if self.use_elution_models:
            pno, nth, peakset, known_peak_info = peakset_info
            paired_ranges_ = [self.cnv_ranges[i] for i in peakset]
            for k, paired_range in enumerate(paired_ranges_):
                if known_peak_info is None:
                    known = False
                else:
                    known = False if known_peak_info[k] is None else True
                for elm_rec in paired_range.elm_recs:
                    e   = elm_rec[0]
                    fnc = elm_rec[1]
                    if fnc.accepts_real_x:
                        # temporary fix under V1PreviewAdapter
                        fy = fnc(self.elution_x)
                    else:
                        fy = fnc(self.fx)
                    if known:
                        ax1.plot( self.elution_x, fy, color='yellow', linewidth=7, alpha=0.3)
                    ax1.plot( self.elution_x, fy, ':', color=get_color(e), linewidth=3 if k == nth else 1.5)
        else:
            if self.doing_sec:
                pass
            else:
                x_ = np.arange(stop - start)
                y_ = self.conc_curves[row](x_)
                ax1.plot(start + x_, y_, ':', color=get_color(row), linewidth=3)

        ymin, ymax = ax1.get_ylim()
        p = mpl_patches.Rectangle(
                (self.j0+start, ymin),  # (x,y)
                stop - start,   # width
                ymax - ymin,    # height
                facecolor   = 'cyan',
                alpha       = 0.2,
            )
        ax1.add_patch( p )

        rank = C.shape[0]
        xmin, xmax = ax1.get_xlim()
        ymin, ymax = ax1.get_ylim()
        tx = xmin * 0.95 + xmax * 0.05
        ty = ymin * 0.3 + ymax * 0.7
        rank_increment = 0 if self.solver.rank_control else self.rank_increment
        rank_inc = '+%d' % rank_increment if rank_increment > 0 else ''
        ax1.text(tx, ty, "Rank: %d%s" % (rank, rank_inc), alpha=0.2, fontsize=20 )

        for ax in [ax2A1]:
            ax.set_xlabel( 'Q' )
            ax.set_ylabel( 'log( I/C )' )
            ax.yaxis.set_major_formatter(plt.FuncFormatter(format_func))

        if need_bq:
            ax2B1.axes.get_yaxis().set_ticks([])
            ax2B1.set_axis_on()
            # ax2B1.set_ylabel( 'I/C' )
        else:
            ax2B1.set_axis_off()

        Y_ = self.Y_list[row]

        logY = np.log10(Y_)
        logA = np.log10(A)

        ax2A1.set_title("Log Plot", fontsize=16)
        ax2A1.plot( self.q, logY, label='ln(I(q))' )
        ax2A1.plot( self.q, logA, label='ln(A(q))' )
        if need_bq:
            ax2B1.plot( self.q, B, color='pink', label='B(q)' )
            if self.lrf_bound_correction:
                try:
                    ymin, ymax = ax2B1.get_ylim()
                    ax2B1.set_ylim(ymin, ymax)
                    for k, bound in enumerate(lrf_info.bq_bounds):
                        label = "B(q) bounds" if k == 0 else None
                        ax2B1.plot(self.q, bound, ":", color='red', label=label)
                except:
                    log_exception(self.logger, "plot bb_bounds failure: ")
            ax2_ = ax2B1
        else:
            ax2_ = ax2A1

        ax2A1.legend()
        if need_bq:
            ax2B1.legend(bbox_to_anchor=(1, 0.1), loc='lower right')
        ax2B1.grid(False)
        if animation:
            return self.solver.get_anim_data()
        else:
            return lrf_info, C

    def plot_guinier_kratky(self, row, devel=False):
        if devel:
            from importlib import reload
            import molass_legacy.Kratky.GuinierKratkyPlots
            reload(molass_legacy.Kratky.GuinierKratkyPlots)
        from molass_legacy.Kratky.GuinierKratkyPlots import guinier_kratky_plots_impl
        ax3, ax4 = self.axis_array[row][3:]
        ay = self.solver_results[row][0]     # A
        sg = self.sg_list[row]
        color = "C1"
        this_rg = sg.Rg
        if this_rg is None or this_rg == 0:
            this_rg = None
            interval = None
        else:
            interval = sg.guinier_start, sg.guinier_stop

        guinier_kratky_plots_impl(ax3, ax4, self.q, ay, this_rg, sg.Iz, color, interval=interval, markersize=1)

        last_rg = self.sg_list[row-1].Rg if row > 0 else None

        rg_text_color = 'black'
        if last_rg is not None and this_rg is not None:
            if self.doing_sec:
                if last_rg < this_rg:
                    rg_text_color = 'red'
            else:
                if last_rg > this_rg:
                    rg_text_color = 'red'

        xmin, xmax = ax3.get_xlim()
        ymin, ymax = ax3.get_ylim()
        tx = xmin * 0.95 + xmax * 0.05
        ty = ymin * 0.8 + ymax * 0.2

        Rg_ = 'None' if this_rg is None else '%.1f' % this_rg
        rg_text = ax3.text(tx, ty, 'Rg: %s' % Rg_, color=rg_text_color, alpha=0.2, fontsize=20 )

    def validate( self ):
        solving_required = False
        for panel in self.panel_list:
            if panel.solving_required:
                solving_required = True
                break
        if solving_required:
            import molass_legacy.KekLib.OurMessageBox as MessageBox
            ok = MessageBox.showinfo( "Confirmation",
                    "You have to 'Cancel' to finish because you have not 'Solve'd with changed parameters.",
                    parent=self )
            return False

        return True

    def apply( self ):  # overrides parent class method
        self.applied = True

        ignore_bq_list = []
        for k, panel in enumerate(self.panel_list):
            self.zx_penalty_matrix[k] = panel.get_penalty_weights()
            ignore_bq_list.append(panel.ignore_bq.get())

        ignore_bq_set = set(ignore_bq_list)
        num_states = len(ignore_bq_set)
        if num_states == 1:
            unique_value = list(ignore_bq_set)[0]
            ignore_all_bqs = unique_value
            if unique_value == 0:
                ignore_bq_list = None
        else:
            ignore_all_bqs = 0

        # when this set is not done, default values are used in ExtrapolationSolver
        set_setting( 'zx_penalty_matrix', self.zx_penalty_matrix )
        set_setting( 'ignore_bq_list', ignore_bq_list )
        set_setting( 'ignore_all_bqs', ignore_all_bqs )
        known_info_list = self.get_known_info_list()
        set_setting('known_info_list', known_info_list)

        print('applied: zx_penalty_matrix=', self.zx_penalty_matrix)
        self.try_image_save()

    def try_image_save( self ):
        test_pattern = get_setting('test_pattern')
        if test_pattern is None or test_pattern < 7:
            return

        preview_image_folder = get_setting('preview_image_folder')
        if preview_image_folder is not None and os.path.exists(preview_image_folder):
            self.save_the_figure( preview_image_folder, get_setting('analysis_name') )

    def save_the_figure( self, folder, analysis_name ):
        from molass_legacy.SerialAnalyzer.DataUtils import cut_upper_folders
        in_folder = cut_upper_folders(get_setting('in_folder'))
        self.fig.suptitle(in_folder)
        self.fig.subplots_adjust(top=0.9)

        # print( 'save_the_figure: ', folder, analysis_name )
        filename = analysis_name.replace( 'analysis', 'figure' )
        path = os.path.join( folder, filename )
        self.fig.savefig( path )

    def get_ignore_bq( self, row ):
        return self.panel_list[row].ignore_bq.get()

    def save_results( self ):
        from .PreviewResultSaver import PreviewResultSaverDialog
        self.saver = PreviewResultSaverDialog(self, self)
        self.saver.show()

    def plot_known(self, row, selected, data):
        from molass_legacy.GuinierAnalyzer.AutorgKekAdapter import AutorgKekAdapter
        # unify with the part of solve_plot

        print('plot_known: row=', row, data.shape)
        ax1, ax2A1, ax2B1 = self.axis_array[row]

        autorg_kek = AutorgKekAdapter( data )
        autorg_result = autorg_kek.run()

        print(autorg_result.Rg)

        for ax in [ax2A1, ax2B1]:
            ax.cla()

        for ax in [ax2A1]:
            ax.set_xlabel( 'Q' )
            ax.set_ylabel( 'log( I/C )' )
            ax.yaxis.set_major_formatter(plt.FuncFormatter(format_func))

        Y_ = self.Y_list[row]
        A = data[:,1]

        logY = np.log10(Y_)
        logA = np.log10(A)

        last_rg = self.sg_list[row-1].Rg if row > 0 else None
        this_rg = autorg_result.Rg

        rg_text_color = 'black'
        if last_rg is not None and this_rg is not None:
            if self.doing_sec:
                if last_rg < this_rg:
                    rg_text_color = 'red'
            else:
                if last_rg > this_rg:
                    rg_text_color = 'red'

        ax2A1.plot( self.q, logY, label='ln(I(q))' )
        ax2A1.plot( self.q, logA, label='ln(A(q))' )

        xmin, xmax = ax2A1.get_xlim()
        ymin, ymax = ax2A1.get_ylim()
        tx = xmin * 0.95 + xmax * 0.05
        ty = ymin * 0.8 + ymax * 0.2
        Rg_ = 'None' if this_rg is None else '%.1f' % autorg_result.Rg
        rg_text = ax2A1.text(tx, ty, 'Rg: %s' % Rg_, color=rg_text_color, alpha=0.2, fontsize=20 )

        ax2A1.legend()
        self.canvas_draw()

    def solve_unknowns(self):
        from .PreviewController import PreviewController

        self.su_btn_blink.stop()

        known_info_list = self.get_known_info_list()

        self.preview_ctl = PreviewController(dialog=self.editor_dialog, editor=self.editor_frame)
        self.preview_ctl.run_solver(self, self.pdata, self.popts, known_info_list=known_info_list)

        if self.preview_ctl.ok():
            self.preview_ctl.show_dialog()
            # self.update_widgets()
            if self.preview_ctl.dialog.applied:
                # self.update_settings()
                # TODO:
                pass
        else:
            # TODO
            pass

    def get_known_info_list(self):
        from .ExtrapolationSolver import KnownInfo
        known_info_list = []
        num_knowns = 0
        for k, panel in enumerate(self.super_panel_list):
            known = panel.known.get()
            print([k], known)
            if known:
                try:
                    info = KnownInfo(panel.known_panel.data)
                    num_knowns += 0
                except:
                    # temp fix for the case when panel.known_panel.data is not ready
                    info = None
            else:
                info = None
            known_info_list.append(info)
        return None if num_knowns == 0 else known_info_list

    def update_to_solve_range(self, row, f, t):
        to_solve_row = self.to_solve_ranges[row]
        to_solve_row[1] = f
        to_solve_row[2] = t

    def auto_cdi_for_all(self, out_folder=None):
        from molass_legacy.Conc.CdInspection import CdInspectionDailog
        print('auto_cdi_for_all')
        if out_folder is None:
            out_folder = get_setting('analysis_folder') + '/cdi-images'
            if not os.path.exists(out_folder):
                from molass_legacy.KekLib.BasicUtils import mkdirs_with_retry
                mkdirs_with_retry(out_folder)

        xray_scale = self.sd.get_xray_scale()
        row = 0
        for pno, paired_range in enumerate(self.cnv_ranges):
            fromto_list = paired_range.get_fromto_list()
            for ad, _ in enumerate(fromto_list):
                ignore_bq = False
                print([pno, ad], paired_range)
                q = self.q
                f, t = paired_range[ad]
                eslice = slice(f, t+1)
                M = self.data[:, eslice]
                E = self.error[:, eslice]
                from_ax = self.axis_array[row][0]
                C = self.panel_list[row].C
                self.logger.info("cdi for peak %d-%d elution range(%d, %d)", pno, ad, f, t+1)
                dialog = CdInspectionDailog(self.parent, self, M, E, C, q, eslice, from_ax=from_ax, xray_scale=xray_scale)
                def auto_action():
                    dialog.save_the_figure(out_folder, pno, ad)
                    dialog.ok()
                dialog.after(1000, auto_action)
                dialog.show()
                row += 1

    def log_rg_info(self):
        rg_list = [sg.Rg for sg in self.sg_list]
        self.logger.info("rg_list=%s", str(rg_list))

    def add_dnd_bind(self):
        self.mpl_canvas_widget.register_drop_target("*")

        def dnd_handler(event):
            self.on_drop(event)

        self.mpl_canvas_widget.bind("<<Drop>>", dnd_handler)

    def on_drop(self, event, debug=True):
        from molass_legacy.SerialAnalyzer.SerialDataUtils import serial_np_loadtxt
        if debug:
            import molass_legacy.SecTools.LogGuinierKratkyPlotter
            from importlib import reload
            reload(molass_legacy.SecTools.LogGuinierKratkyPlotter)
        from molass_legacy.SecTools.LogGuinierKratkyPlotter import LogGuinierKratkyPlotter

        files = event.data.split(' ')
        print('on_drop:', files)

        path = files[0]
        data, _ = serial_np_loadtxt(path)
        print("data.shape=", data.shape)
        plotter = LogGuinierKratkyPlotter(self, data, path)
        plotter.show()

    def on_the_fly(self):
        from OnTheFly.DebugDialog import DebugDialog
        dialog = DebugDialog(self)
        dialog.show()



# coding: utf-8
"""
    SvdTutorial2D.py

    Copyright (c) 2019,2024, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.patches import Polygon, Circle
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import tk_set_icon_portable
import molass_legacy.KekLib.DebugPlot as dplt
from molass_legacy.KekLib.OurMatplotlib import get_color
from ModeledData import ModeledData
from MatrixData import simple_plot_3d
from .SvdPrecision import plot_text

FONTSIZE = 200

class SvdTutorial2D(Dialog):
    def __init__(self, parent):
        self.parent = parent
        Dialog.__init__(self, parent, "SVD Tutorial", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        tk_set_icon_portable(self)

        cframe = Tk.Frame(body_frame)
        cframe.pack()

        self.fig = fig = plt.figure(figsize=(15,10))
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)
        self.mpl_canvas.mpl_connect('button_press_event', self.on_press)
        self.create_axes()
        self.draw()
        fig.tight_layout()
        self.mpl_canvas.draw()

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Animate", width=10, command=self.animate, state=Tk.DISABLED)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

    def create_axes(self):
        fig = self.fig
        gs = GridSpec(4,3)
        ax1 = fig.add_subplot(gs[1:3, 0])
        ax2 = fig.add_subplot(gs[0:2, 1])
        ax3 = fig.add_subplot(gs[1:3, 2])
        ax4 = fig.add_subplot(gs[2:4, 1])
        self.axes = [ax1, ax2, ax3, ax4]
        self.ax5 = fig.add_subplot(gs[3, 2])
        self.ax5.set_axis_off()
        plot_text(self.ax5, "Click anywhere to play", alpha=0.2, fontsize=30)

    def draw(self):
        self.generate_random_points()
        ax1, ax2, ax3, ax4 = self.axes
        self.draw1(ax1)
        self.draw2(ax2)
        self.draw3(ax3)
        self.draw4(ax4)

    def on_press(self, event):
        for ax in self.axes:
            ax.cla()
        self.draw()
        self.mpl_canvas.draw()

    def generate_random_points(self):
        a = np.random.rand()*2 - 1
        b = np.random.rand()*2 - 1
        c = np.random.rand()*2 - 1
        d = np.random.rand()*2 - 1
        self.points = [(a,b), (c,d)]
        self.M = np.array(self.points).T
        U, s, VT = np.linalg.svd(self.M)
        self.U = U
        self.S = np.diag(s)
        self.V = VT.T

    def draw_points(self, ax, W):
        px = W[0,:]
        py = W[1, :]
        vx = np.sum(px)
        vy = np.sum(py)
        p = W[:,0]
        q = W[:,1]
        ax.plot(*p, 'o')
        ax.plot(*q, 'o')
        poly_points = [(0,0), p, (vx,vy), q]
        polygon = Polygon(poly_points, alpha=0.5, color='cyan')
        ax.add_patch(polygon)
        poly_points = [(0,0), (1,0), (1,1), (0,1)]
        polygon = Polygon(poly_points, alpha=0.1, color='red')
        ax.add_patch(polygon)

    def draw_disc(self, ax):
        circle = Circle((0,0), 1, alpha=0.1)
        ax.add_patch(circle)

    def draw1(self, ax):
        self.draw_points(ax, self.M)
        ax.set_xlim(-2, 2)
        ax.set_ylim(-2, 2)
        plot_text(ax, "$M$", fontsize=FONTSIZE)

    def draw2(self, ax):
        self.draw_points(ax, self.U)
        ax.set_xlim(-2, 2)
        ax.set_ylim(-2, 2)
        self.draw_disc(ax)
        plot_text(ax, "$V^t$", fontsize=FONTSIZE)

    def draw3(self, ax):
        self.draw_points(ax, self.S)
        ax.set_xlim(-2, 2)
        ax.set_ylim(-2, 2)
        plot_text(ax, "$\Sigma$", fontsize=FONTSIZE)

    def draw4(self, ax):
        self.draw_points(ax, self.V)
        self.draw_disc(ax)
        ax.set_xlim(-2, 2)
        ax.set_ylim(-2, 2)
        plot_text(ax, "$U$", fontsize=FONTSIZE)

    def animate(self):
        self.show_matrix_mapping()
        self.make_anim_VT()
        self.make_anim_S()
        self.make_anim_U()
        self.show_anim()

    def show_matrix_mapping(self):
        dplt.push()
        fig, axes = dplt.subplots(nrows=1, ncols=2, figsize=(12,6))
        ax1, ax2 = axes
        ax1.plot(0,0,'o')
        ax2.set_axis_off()
        fig.tight_layout()
        dplt.show()
        dplt.pop()

    def make_anim_VT(self):
        pass

    def make_anim_S(self):
        pass

    def make_anim_U(self):
        pass

    def show_anim(self):
        pass



# coding: utf-8
"""

    SerialAnalyzer.TesterMiscellaneous.py

    Copyright (c) 2018-2020, SAXS Team, KEK-PF

"""

import os
import re
import queue
import threading
import numpy        as np
from time           import sleep
from molass_legacy.KekLib.OurTkinter     import Tk, is_empty_val
from DataUtils      import serial_folder_walk
from TkTester       import TestClient, reply_ok, reply_yes, reply_no, messagebox_shown
from molass_legacy.KekLib.BasicUtils     import clear_dirs_with_retry, mkdirs_with_retry, open_w_safely
from molass_legacy._MOLASS.SerialSettings import clear_settings, get_setting, set_setting
from molass_legacy.KekLib.ExceptionTracebacker       import ExceptionTracebacker
from LightObjects   import AnyObject

def normal_input( agent, self ):
        print( 'in_folder=', self.in_folder )

        agent.in_folder_entry.focus_force()
        agent.in_folder_entry.delete( 0, Tk.END)
        agent.in_folder_entry.insert( 0, self.in_folder )
        agent.on_entry_in_folder()

        while is_empty_val(agent.uv_folder.get()):
            print('waiting for uv_folder update')
            # agent.update()
            sleep(1)

        agent.an_folder_entry.focus_force()
        agent.an_folder_entry.delete( 0, Tk.END)
        agent.an_folder_entry.insert( 0, self.an_folder )
        agent.on_entry_an_folder()
        sleep(1)

class NormalInput:
    def __init__( self, app ):
        self.app    = app

    def test_folders( self, in_folders, an_folder ):
        self.in_folders  = in_folders
        self.an_folder  = an_folder
        self.client = TestClient( self.app, self.test_func )

    def test_func( self, client, agent ):
        self.agent  = agent

        for folder in self.in_folders:
            self.in_folder = folder
            normal_input( agent, self )
            sleep( 5 )

        agent.quit( immediately=True, __wait__=False )
        print( 'agent.quit' )

class VariationBase:
    def do_assert( self, assert_key, value ):
        assert_ = self.assert_dict.get( assert_key )
        if assert_ is None:
            print( 'no assert for ' + assert_key )
        else:
            try:
                assert_( value )
                print( assert_key + ' assert ok' )
            except:
                etb = ExceptionTracebacker()
                print( etb )
                print( assert_key + ' assert failed' )

class MapperCanvasVariation1(VariationBase):
    def __init__( self, app ):
        self.app    = app

    def test_a_folder( self, in_folder, an_folder, assert_dict ):
        self.in_folder  = in_folder
        self.an_folder  = an_folder
        self.assert_dict    = assert_dict
        self.client = TestClient( self.app, self.test_func )

    def test_func( self, client, agent ):
        self.agent  = agent

        normal_input( agent, self )

        # ---- guinier animation
        agent.file_info_table.table.select_row( 141 )
        agent.file_info_table.run_simple_guinier_animation( __wait__=False )
        sleep( 5 )
        agent.file_info_table.anim_dialog.ok_button.invoke()

        # ---- default manipulation
        agent.analysis_button.invoke( __wait__=False )
        sleep( 1 )
        analyzer_dialog = agent.analyzer.dialog
        mapper_canvas  = analyzer_dialog.mapper_canvas
        while True:
            try:
                mapper_canvas.ok_button.invoke()
                break
            except:
                pass
            sleep( 1 )

        analyzer_dialog.cancel_button.invoke()
        sleep( 0.5 )
        reply_ok()

        range_type = agent.get_range_type()
        self.do_assert( 'ranges1-0', range_type )

        ranges1 = agent.analyzer.dialog.get_analysis_ranges_old_style()
        self.do_assert( 'ranges1-1', ranges1 )

        # ---- range adjustment
        agent.analysis_button.invoke( __wait__=False )
        sleep( 1 )
        mapper_canvas  = analyzer_dialog.mapper_canvas
        mapper_canvas.reditor_btn.invoke()
        sleep( 1 )

        for k in range(2):
            print( k, 'buttonup' )
            mapper_canvas.range_list_entry.range_entries[0][0].invoke( 'buttonup' )
            sleep( 0.1 )
        sleep( 1 )

        mapper_canvas.ok_button.invoke()
        sleep( 1 )
        analyzer_dialog.cancel_button.invoke()
        sleep( 0.5 )
        reply_ok()

        range_type = agent.get_range_type()
        self.do_assert( 'ranges2-0', range_type )

        ranges2 = agent.analyzer.dialog.get_analysis_ranges_old_style()
        self.do_assert( 'ranges2-1', ranges2 )

        sleep( 3 )
        agent.quit( immediately=True, __wait__=False )
        print( 'agent.quit' )


class MapperCanvasVariation2(VariationBase):
    def __init__( self, app ):
        self.app    = app

    def test_a_folder( self, in_folder, an_folder, assert_dict ):
        self.in_folder  = in_folder
        self.an_folder  = an_folder
        self.assert_dict    = assert_dict
        self.client = TestClient( self.app, self.test_func )

    def test_func( self, client, agent ):
        self.agent  = agent

        normal_input( agent, self )

        agent.analysis_button.invoke( __wait__=False )
        sleep( 1 )

        analyzer_dialog = agent.analyzer.dialog
        mapper_canvas   = analyzer_dialog.mapper_canvas
        mapper_adjuster = None

        while mapper_adjuster is None:
            try:
                mapper_adjuster  = mapper_canvas.adjuster
            except Exception as exc:
                print( exc )
                pass
            sleep( 1 )

        # ---- mapping helper input
        if False:
            mapper_adjuster.helper_btn.invoke( __wait__=False )
            sleep( 1 )

            helper_dialog = analyzer_dialog.helper_dialog
            helper_dialog.flow_change_cb_vars[0].set( 1 )

            axes = helper_dialog.get_axes()

            event = AnyObject( dblclick=True, inaxes=axes[0], xdata=435.0, ydata=0.08 )
            helper_dialog.a_manipulator.modify_peaks( event )

            event = AnyObject( dblclick=True, inaxes=axes[1], xdata=217.0, ydata=0.044 )
            helper_dialog.x_manipulator.modify_peaks( event )

            sleep( 1 )
            helper_dialog.ok_button.invoke()

        # ---- set adjustment options
        mapper_adjuster.uv_baseline_adjust.set( 1 )
        mapper_adjuster.xray_baseline_opt.set( 1 )
        mapper_adjuster.xray_baseline_adjust.set( 1 )
        sleep( 1 )

        # ---- optimize
        mapper_adjuster.optimize_btn.invoke()

        # ---- show 3D view
        mapper_canvas.show_absorbance_button.invoke( __wait__=False  )
        sleep( 1 )
        mapper_canvas.viewer.ok_button.invoke()

        mapper_canvas.show_scattering_button.invoke( __wait__=False  )
        sleep( 2.5 )

        scatterning_plot = mapper_canvas.scatterning_plot
        scatterning_plot.plot_frames[0].vsa_only_cb.invoke()
        scatterning_plot.plot_frames[0].redraw_btn.invoke()

        sleep( 1 )
        scatterning_plot.ok_button.invoke()

        # ---- range adjustment
        mapper_canvas.reditor_btn.invoke()
        sleep( 1 )

        ranges1 = mapper_canvas.range_list_entry.get_ranges()
        self.do_assert( 'ranges1-1', ranges1 )

        mapper_canvas.range_list_entry.range_entries[1][0].invoke( 'buttonup' )
        mapper_canvas.range_list_entry.range_entries[1][2].invoke( 'buttondown' )

        ranges2 = mapper_canvas.range_list_entry.get_ranges()
        self.do_assert( 'ranges2-1', ranges2 )

        sleep( 1 )
        mapper_canvas.ok_button.invoke()
        sleep( 1 )
        analyzer_dialog.cancel_button.invoke()
        sleep( 0.5 )
        reply_ok()

        range_type = agent.get_range_type()
        self.do_assert( 'ranges2-0', range_type )

        ranges2 = agent.analyzer.dialog.get_analysis_ranges_old_style()
        self.do_assert( 'ranges2-1', ranges2 )

        sleep( 1 )
        agent.quit( immediately=True, __wait__=False )
